/*
 *  @(#) $Id: boxcount.c,v 1.11 2004/05/13 18:11:34 yeti Exp $
 *  Copyright (C) 2004 David Necas (Yeti)
 *  E-mail: yeti@physics.muni.cz.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA
 */

#include "ifsgr.h"
#include <string.h>
#include <stdlib.h>
#include <math.h>

static inline uint32_t
count_nonzero(size_t rows,
              size_t cols,
              uint32_t *canvas)
{
    size_t i, j;
    uint32_t m;

    m = 0;
    for (i = 0; i < rows; i++) {
        size_t *row = canvas + i*cols;

        for (j = cols; j; j--, row++)
            m += (*row != 0);
    }

    return m;
}

static inline uint32_t
count_nonzero_2x2(size_t rows,
                  size_t cols,
                  uint32_t *canvas)
{
    size_t i, j;
    uint32_t m;

    m = 0;
    for (i = 0; i < rows/2; i++) {
        size_t *row = canvas + (2*i)*cols;
        size_t *row2 = canvas + (2*i+1)*cols;

        for (j = cols/2; j; j--, row += 2, row2 += 2)
            m += *row || row[1] || *row2 || row2[1];
    }

    return m;
}

/**
 * ifs_boxcount:
 * @rows: The number of @canvas rows.
 * @cols: The number of @canvas columns.
 * @canvas: A rendered fractal @rows x @cols, nonzeros are points of the
 *          fractal.
 * @maxsize: Maximum box size (the returned array will be of this size).
 *
 * Count the number of boxes needed to cover the fractal in @canvas.
 *
 * Returns: The box counts, starting from boxes of size 1.
 **/
uint32_t*
ifs_boxcount(size_t rows,
             size_t cols,
             uint32_t *canvas,
             size_t maxsize)
{
    size_t i, j, ii, jj, s, m;
    size_t *counts;

    if (!maxsize)
        return NULL;
    counts = (uint32_t*)calloc(maxsize, sizeof(uint32_t));
    counts[0] = count_nonzero(rows, cols, canvas);
    if (maxsize < 2)
        return counts;
    counts[1] = count_nonzero_2x2(rows, cols, canvas);
    if (maxsize < 3)
        return counts;

    for (m = 3; m <= maxsize; m++) {
        s = 0;
        for (i = 0; i < (rows+m-1)/m; i++) {
            for (j = 0; j < (cols+m-1)/m; j++) {
                for (ii = 0; ii < m; ii++) {
                    if (i*m+ii >= rows)
                        break;
                    for (jj = 0; jj < m; jj++) {
                        if (j*m+j >= cols)
                            break;
                        if (canvas[(i*m+ii)*cols + j*m+jj]) {
                            s++;
                            ii = m;  /* break out of the outer cycle */
                            break;
                        }
                    }
                }
            }
        }
        counts[m-1] = s;
    }
    return counts;
}

/**
 * ifs_estimate_dimension:
 * @ifs: An Iterated Function System.
 * @ntab: The size of @ntab.
 * @tab: A precomputed transformation relative occurence table.
 *
 * Quickly estimates the box-counting dimension of @ifs.
 *
 * Returns: The box-counting dimension of @ifs.
 **/
double
ifs_estimate_dimension(const IFS *ifs,
                       size_t ntab,
                       const size_t *tab)
{
    double dim;

    ifs_estimate_dimension2(ifs, ntab, tab, &dim, NULL);
    return dim;
}

/**
 * ifs_estimate_dimension2:
 * @ifs: An Iterated Function System.
 * @ntab: The size of @ntab.
 * @tab: A precomputed transformation relative occurence table.
 * @dim: Where the estimated dimension should be stored, if not %NULL.
 * @measure: Where the estimated measure should be stored, if not %NULL.
 *
 * Quickly estimates the box-counting dimension and measure of @ifs.
 **/
void
ifs_estimate_dimension2(const IFS *ifs,
                        size_t ntab,
                        const size_t *tab,
                        double *dim,
                        double *measure)
{
    static const size_t size = 120;
    static const size_t max = 12;
    uint32_t *canvas = NULL;
    size_t *counts;
    uint64_t niter;

    if (!dim && !measure)
        return;
    canvas = (uint32_t*)malloc(size*size*sizeof(uint32_t));
    niter = (uint64_t)exp(3.5+1.6*log(size));
    memset(canvas, 0, size*size*sizeof(uint32_t));
    ifs_draw_int(ifs, ntab, tab, niter, size, size, canvas);
    counts = ifs_boxcount(size, size, canvas, max);
    free(canvas);
    ifs_fit_boxcount_dimension2(max, counts, dim, measure);
    free(counts);
}

/**
 * ifs_fit_boxcount_dimension:
 * @max: The maximum box size (and size of @counts).
 * @counts: The box counts, starting from size 1 (as returned by
 *          ifs_boxcount()).
 *
 * Estimates the box-counting dimension from box counts in @counts.
 *
 * Returns: The estimated dimension.
 **/
double
ifs_fit_boxcount_dimension(size_t max,
                           const size_t *counts)
{
    double dim;

    ifs_fit_boxcount_dimension2(max, counts, &dim, NULL);
    return dim;
}

/**
 * ifs_fit_boxcount_dimension2:
 * @max: The maximum box size (and size of @counts).
 * @counts: The box counts, starting from size 1 (as returned by
 *          ifs_boxcount()).
 * @dim: Where the estimated dimension should be stored, if not %NULL.
 * @measure: Where the estimated measure should be stored, if not %NULL.
 *
 * Estimates the box-counting dimension and measure from box counts in @counts.
 **/
void
ifs_fit_boxcount_dimension2(size_t max,
                            const size_t *counts,
                            double *dim,
                            double *measure)
{
    double sumx, sumxx, sumy, sumxy;
    size_t i;

    if (!dim && !measure)
        return;

    sumx = sumy = sumxx = sumxy = 0.0;
    for (i = 0; i < max; i++) {
        double x = -log((i+1.0)/50.0);
        double y = log(1.0*counts[i]);

        sumx += x;
        sumy += y;
        sumxx += x*x;
        sumxy += x*y;
    }

    if (dim)
        *dim = (sumxy - sumx*sumy/max)/(sumxx - sumx*sumx/max);
    if (measure)
        *measure = (sumy*sumxx - sumxy*sumx)/(sumxx - sumx*sumx/max);
}

/* vim: set cin et ts=4 sw=4 cino=>1s,e0,n0,f0,{0,}0,^0,\:1s,=0,g1s,h0,t0,+1s,c3,(0,u0 : */
