/*
 *  @(#) $Id: draw.c,v 1.21 2004/05/13 18:11:34 yeti Exp $
 *  Copyright (C) 2004 David Necas (Yeti)
 *  E-mail: yeti@physics.muni.cz.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA
 */

#include "ifsgr.h"
#include <stdlib.h>
#include <string.h>
#include <math.h>

/**
 * ifs_draw_int:
 * @ifs: An iterated function system.
 * @ntab: The size of @ntab.
 * @tab: A precomputed transformation relative occurence table.
 * @niter: The number of iterations to perform.
 * @rows: Height of canvas.
 * @cols: Width of canvas.
 * @canvas: 32bit-integer canvas to draw pixel hit counts to.
 *
 * Render an iterated function system @ifs to @canvas.
 *
 * On returns the canvas contains the numbers how many times each pixel was hit
 * during the iterations.
 **/
void
ifs_draw_int(const IFS *ifs,
             size_t ntab,
             const size_t *tab,
             uint64_t niter,
             size_t rows,
             size_t cols,
             uint32_t *canvas)
{
    uint64_t i;
    size_t row, col;
    double x, y, xx;

    x = y = 0.0;
    memset(canvas, 0, sizeof(size_t)*rows*cols);

    /* get to the attractor */
    for (i = 0; i < 12; i++) {
        IFSTrans *t = ifs->t + tab[ifs_taus_get() % ntab];
        t = ifs->t + i % ifs->n;
        xx = x;
        x = xx*t->xx + y*t->xy + t->x;
        y = xx*t->yx + y*t->yy + t->y;
    }

    /* go! */
    i = niter;
    while (i--) {
        IFSTrans *t = ifs->t + tab[ifs_taus_get() % ntab];
        xx = x;
        x = xx*t->xx + y*t->xy + t->x;
        y = xx*t->yx + y*t->yy + t->y;

        row = rows*y;
        if (row >= rows)
            continue;
        col = cols*x;
        if (col >= cols)
            continue;
        canvas[row*cols + col]++;
    }
}

/**
 * ifs_int_get_max:
 * @rows: Height of canvas.
 * @cols: Width of canvas.
 * @canvas: 32bit-integer canvas with pixel hit counts.
 *
 * Find the maximum pixel hit count.
 *
 * Returns: The maximum pixel hit count.
 **/
uint32_t
ifs_int_get_max(size_t rows,
                size_t cols,
                const uint32_t *canvas)
{
    size_t i;
    uint32_t m;

    m = 0;
    for (i = rows*cols; i; i--, canvas++) {
        if (*canvas > m)
            m = *canvas;
    }

    return m;
}

/**
 * ifs_int_get_sqavg:
 * @rows: Height of canvas.
 * @cols: Width of canvas.
 * @canvas: 32bit-integer canvas with pixel hit counts.
 *
 * Compute average of squared pixel hit counts.
 *
 * Returns: The average of pixel hit count squares.
 **/
double
ifs_int_get_sqavg(size_t rows,
                  size_t cols,
                  const uint32_t *canvas)
{
    size_t i;
    double sum = 0;

    for (i = rows*cols; i; i--, canvas++)
        sum += (double)*canvas*(*canvas);
    return sum/(rows*cols);
}

/**
 * ifs_int_get_histogram:
 * @rows: Height of canvas.
 * @cols: Width of canvas.
 * @canvas: 32bit-integer canvas with pixel hit counts.
 * @m: Histogram size (the maximum pixel hit count).
 *
 * Create a histogram of pixel hit counts.
 *
 * Note @m has to be greater or equal to the real maximum pixel hit count,
 * no boundary checking is done.
 *
 * Returns: A newly allocated array of size @m+1 with the pixel hit count
 *          histogram.
 **/
uint32_t*
ifs_int_get_histogram(size_t rows,
                      size_t cols,
                      const uint32_t *canvas,
                      uint32_t m)
{
    uint32_t *hist;
    size_t i;

    hist = (uint32_t*)calloc(m+1, sizeof(uint32_t));
    for (i = rows*cols; i; i--, canvas++) {
        hist[*canvas]++;
    }

    return hist;
}

/**
 * ifs_int_get_nonzero:
 * @rows: Height of canvas.
 * @cols: Width of canvas.
 * @canvas: 32bit integer canvas with pixel hit counts.
 *
 * Count pixels with nonzero hit count.
 *
 * Returns: The number of pixels with nonzero hit count in @canvas.
 **/
uint32_t
ifs_int_get_nonzero(size_t rows,
                    size_t cols,
                    const uint32_t *canvas)
{
    size_t i;
    uint32_t m;

    m = 0;
    for (i = rows*cols; i; i--, canvas++) {
        if (*canvas)
            m++;
    }

    return m;
}

/**
 * ifs_int2gray8_bw:
 * @pixmap: Grayscale 1 byte per pixel pixmap to render @canvas to.
 * @canvas: 32bit integer canvas with pixel hit counts.
 * @rows: Height of canvas.
 * @cols: Width of canvas.
 * @threshold: Black-white threshold, pixels with lower hit counts than
 *             @threshold are drawn black (0) others are drawn white (255).
 *
 * Render pixel hit counts as black-and-white image.
 **/
void
ifs_int2gray8_bw(unsigned char *pixmap,
                 const uint32_t *canvas,
                 size_t rows,
                 size_t cols,
                 uint32_t threshold)
{
    size_t i, j;

    for (i = 0; i < rows; i++) {
        const uint32_t *crow = canvas + i*cols;
        unsigned char *prow = pixmap + i*cols;

        for (j = 0; j < cols; j++) {
            prow[j] = crow[j] > threshold ? 0 : 255;
        }
    }
}

/**
 * ifs_int2gray8_converg:
 * @pixmap: Grayscale 1 byte per pixel pixmap to render @canvas to.
 * @canvas: 32bit integer canvas with pixel hit counts.
 * @rows: Height of canvas.
 * @cols: Width of canvas.
 * @max: The maximum pixel hit count.
 * @qgamma: Gamma value, value 1.0 means linear mapping.  Usual gamma values
 *          lie in wide range approximately from 1/10 to 1000.
 *          See the TeX files in the source distribution for some discussion
 *          of good gamma choice.
 *
 * Render pixel hit counts as grayscale image with 1 byte per pixel.
 **/
void
ifs_int2gray8_converg(unsigned char *pixmap,
                      const uint32_t *canvas,
                      size_t rows,
                      size_t cols,
                      uint32_t max,
                      double qgamma)
{
    size_t i, j;

    for (i = 0; i < rows; i++) {
        const uint32_t *crow = canvas + i*cols;
        unsigned char *prow = pixmap + i*cols;

        for (j = 0; j < cols; j++) {
            double x = crow[j];

            x = exp(log(1.0 + IFS_EPSILON - x/max)*qgamma);
            prow[j] = floor(x >= 1.0 ? 255.0 : 255.999999*x);
        }
    }
}

/**
 * ifs_int2gray16_converg:
 * @pixmap: Grayscale 2 bytes per pixel pixmap to render @canvas to.
 * @canvas: 32bit integer canvas with pixel hit counts.
 * @rows: Height of canvas.
 * @cols: Width of canvas.
 * @max: The maximum pixel hit count.
 * @qgamma: Gamma value, value 1.0 means linear mapping.  Usual gamma values
 *          lie in wide range approximately from 1/10 to 1000.
 *          See the TeX files in the source distribution for some discussion
 *          of good gamma choice.
 *
 * Render pixel hit counts as grayscale image with 2 bytes per pixel.
 **/
void
ifs_int2gray16_converg(uint16_t *pixmap,
                       const uint32_t *canvas,
                       size_t rows,
                       size_t cols,
                       uint32_t max,
                       double qgamma)
{
    size_t i, j;

    for (i = 0; i < rows; i++) {
        const uint32_t *crow = canvas + i*cols;
        uint16_t *prow = pixmap + i*cols;

        for (j = 0; j < cols; j++) {
            double x = crow[j];

            x = exp(log(1.0 + IFS_EPSILON - x/max)*qgamma);
            prow[j] = floor(x >= 1.0 ? 255.0 : 65535.9999*x);
        }
    }
}

/**
 * ifs_discretize_probability:
 * @ifs: An Iterated Function System.
 * @ntab: The size of @ntab.
 * @tab: Array to store the precomputed transformation relative occurrence
 *       table to, or %NULL to allocate a new one.
 *
 * Computes transformation relative occurrence table.
 *
 * Index of each transformation is present proportionally to the transformation
 * probability.  To select a random transformation with corresponding
 * probability is thus sufficient to select a random item from the table.
 *
 * Returns: @tab itself it it was not %NULL, a newly allocated array which
 *          should be freed when no longer needed otherwise.
 **/
size_t*
ifs_discretize_probability(const IFS *ifs,
                           size_t ntab,
                           size_t *tab)
{
    ssize_t i, j, m;
    double s;

    if (!tab)
        tab = (size_t*)malloc(ntab*sizeof(size_t));
    j = m = -1;
    s = 0.0;
    for (i = 0; i < (ssize_t)ntab; i++) {
        if (i > m) {
            j++;
            s += ifs->t[j].p;
            m = ntab*s;
        }
        tab[i] = j;
    }

    return tab;
}

/*********** Documentation ***************************************************/
/**
 * IFS_PROBSTACK_BITS:
 *
 * The number of bits needed to represent indices in precomputed probability
 * maps used in IFS drawing functions.
 **/

/**
 * IFS_PROBSTACK:
 *
 * The size of precomputed probability maps used in IFS drawing functions.
 **/

/**
 * IFS_PROBSTACK_MASK:
 *
 * Mask for clearing all but the lowest @IFS_PROBSTACK_BITS bits in an integer.
 **/

/* vim: set cin et ts=4 sw=4 cino=>1s,e0,n0,f0,{0,}0,^0,\:1s,=0,g1s,h0,t0,+1s,c3,(0,u0 : */
