/*
 *  @(#) $Id: file.c,v 1.22 2004/04/06 12:11:11 yeti Exp $
 *  Copyright (C) 2004 David Necas (Yeti)
 *  E-mail: yeti@physics.muni.cz.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA
 */

#include "ifsgr.h"
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>

enum {
    /* element */
    IFSC_KEY_X,
    IFSC_KEY_Y,
    IFSC_KEY_THETA,
    IFSC_KEY_SCALE,
    IFSC_KEY_ASYM,
    IFSC_KEY_SHEAR,
    IFSC_KEY_FLIP,
    IFSC_KEY_RED_COLOR,
    IFSC_KEY_GREEN_COLOR,
    IFSC_KEY_BLUE_COLOR,
    IFSC_KEY_BLACK_COLOR,
    IFSC_KEY_TARGET_COLOR,
    IFSC_KEY_HUE_SCALE,
    IFSC_KEY_VALUE_SCALE,
    IFSC_KEY_SIMPLE_COLOR,
    IFSC_KEY_PROB,
    /* global */
    IFSC_KEY_ELEMENT,
    IFSC_KEY_ITERATIONS,
    IFSC_KEY_MAX_MEMORY,
    IFSC_KEY_SUBDIVIDE,
    IFSC_KEY_RADIUS,
    IFSC_KEY_ASPECT_RATIO,
    IFSC_KEY_CENTER_X,
    IFSC_KEY_CENTER_Y,

    IFSC_LAST_KEYWORD
};

/* should be initialized to zeroes */
typedef struct {
    char *buffer;  /* current line */
    char *p;  /* position in buffer */
    size_t len;  /* buffer length */
    char *tok;  /* current token */
} IFSFilePos;

static struct {
    const char *name;
    int key;
}
const IFSC_KEYWORDS[] = {
    /* element */
    { "x", IFSC_KEY_X },
    { "y", IFSC_KEY_Y },
    { "theta", IFSC_KEY_THETA },
    { "scale", IFSC_KEY_SCALE },
    { "asym", IFSC_KEY_ASYM },
    { "shear", IFSC_KEY_SHEAR },
    { "flip", IFSC_KEY_FLIP },
    { "red_color", IFSC_KEY_RED_COLOR },
    { "green_color", IFSC_KEY_GREEN_COLOR },
    { "blue_color", IFSC_KEY_BLUE_COLOR },
    { "black_color", IFSC_KEY_BLACK_COLOR },
    { "target_color", IFSC_KEY_TARGET_COLOR },
    { "hue_scale", IFSC_KEY_HUE_SCALE },
    { "value_scale", IFSC_KEY_VALUE_SCALE },
    { "simple_color", IFSC_KEY_SIMPLE_COLOR },
    { "prob", IFSC_KEY_PROB },
    /* global */
    { "element", IFSC_KEY_ELEMENT },
    { "iterations", IFSC_KEY_ITERATIONS },
    { "max_memory", IFSC_KEY_MAX_MEMORY },
    { "subdivide", IFSC_KEY_SUBDIVIDE },
    { "radius", IFSC_KEY_RADIUS },
    { "aspect_ratio", IFSC_KEY_ASPECT_RATIO },
    { "center_x", IFSC_KEY_CENTER_X },
    { "center_y", IFSC_KEY_CENTER_Y },

    { NULL, IFSC_LAST_KEYWORD }
};

static const char *ifs_delim = " \t\v\n\r\f";
static const char ifs_comment1 = ';';
static const char ifs_comment2 = '#';

static bool        read_transform      (FILE *fh,
                                        IFSFilePos *pos,
                                        IFS *ifs);
static int         get_ifsc_keyword    (FILE *fh,
                                        IFSFilePos *pos);
static const char* get_token           (FILE *fh,
                                        IFSFilePos *pos);
static bool        get_double          (FILE *fh,
                                        IFSFilePos *pos,
                                        double *dbl);
static bool        get_and_ditch_tuple (size_t n,
                                        FILE *fh,
                                        IFSFilePos *pos);
__attribute__((unused))
static const char* ifs_file_pos_rewind (IFSFilePos *pos);
static void        ifs_file_pos_free   (IFSFilePos *pos);
static bool        read_triangle       (FILE *fh,
                                        double *triangle,
                                        bool old);
static void        ifs_trans_from_triangle (const double *trn0,
                                            const double *trn,
                                            IFSTrans *t);
static void        triangle_to_coord   (const double *t,
                                        const double *rhs,
                                        double *x,
                                        double *y,
                                        double *c);

/**
 * ifs_read_next:
 * @fh: A stdio filehandle open for read.
 *
 * Read next 2D IFS from Fractint IFS file.
 *
 * Returns: A newly allocated IFS on success, %NULL on failure.
 **/
IFS*
ifs_read_next(FILE *fh)
{
    return ifs_read_by_name(fh, NULL);
}

/**
 * ifs_read_by_name:
 * @fh: A stdio filehandle open for read.
 * @name: Name of the IFS to find, case-insensitive.
 *
 * Read one complete Fractint 2D IFS from filehandle @fh.
 *
 * Returns: A newly allocated IFS on success, %NULL on failure.
 **/
IFS*
ifs_read_by_name(FILE *fh,
                 const char *name)
{
    IFS *ifs;
    IFSFilePos *pos;
    const char *tok;

    /* find start of a new 2D fractal */
    pos = (IFSFilePos*)calloc(1, sizeof(IFSFilePos));
    do {
        if (!(tok = get_token(fh, pos))) {
            ifs_file_pos_free(pos);
            return NULL;
        }
        if (strcmp(get_token(fh, pos), "{") == 0
            && (!name || !*name || strcasecmp(tok, name) == 0))
            break;
        /* something strange, maybe 3D, skip it */
        while (strcmp(tok = get_token(fh, pos), "}") != 0)
            ;
    } while (true);

    ifs = (IFS*)malloc(sizeof(IFS));
    ifs->name = strdup(tok);
    if (!read_transform(fh, pos, ifs)) {
        free(ifs);
        ifs = NULL;
    }
    ifs_file_pos_free(pos);

    return ifs;
}

/**
 * ifs_read_file:
 * @fh: A stdio filehandle open for read.
 * @nifs: Where the number of reads IFSes should be stored.
 *
 * Read all 2D IFSes from Fractint IFS file.
 *
 * Returns: A newly allocated array of @nifs IFSes.  If nothing is read,
 *          %NULL is returned.
 **/
IFS**
ifs_read_file(FILE *fh,
              size_t *nifs)
{
    IFS *ifs, **all;
    struct stat st;
    size_t n, size = 0;

    if (fstat(fileno(fh), &st))
        size = st.st_size/250;  /* an emprirical average IFS size */
    size = (size > 16) ? size : 16;
    all = (IFS**)calloc(size, sizeof(IFS*));
    n = 0;
    while ((ifs = ifs_read_next(fh))) {
        if (n == size) {
            size *= 2;
            all = (IFS**)realloc(all, size*sizeof(IFS*));
        }
        all[n] = ifs;
        n++;
    }

    all = (IFS**)realloc(all, n*sizeof(IFS*));
    *nifs = n;

    return n ? all : NULL;
}

static bool
read_transform(FILE *fh,
               IFSFilePos *pos,
               IFS *ifs)
{
    const char *tok;
    size_t i;

    ifs->n = 0;
    ifs->t = NULL;
    /* read the transform */
    while ((tok = get_token(fh, pos)) && strcmp(tok, "}") != 0) {
        IFSTrans *t;

        (ifs->n)++;
        /* FIXME: inefficient */
        ifs->t = (IFSTrans*)realloc(ifs->t, ifs->n*sizeof(IFSTrans));
        t = ifs->t + (ifs->n - 1);
        t->xx = atof(tok);
        if (!get_double(fh, pos, &t->xy)
            || !get_double(fh, pos, &t->yx)
            || !get_double(fh, pos, &t->yy)
            || !get_double(fh, pos, &t->x)
            || !get_double(fh, pos, &t->y)
            || !get_double(fh, pos, &t->p)) {
            free(ifs->t);
            return false;
        }
    }

    /* fix the last probability */
    {
        double sum = 0.0;

        for (i = 0; i < ifs->n-1; i++)
            sum += ifs->t[i].p;

        ifs->t[ifs->n - 1].p = 1.0 + IFS_EPSILON - sum;
    }

    return true;
}

/**
 * ifs_write_ifs:
 * @ifs: An iterated function system.
 * @fh: A file handle open for writing to write the IFS to.
 *
 * Write @ifs in Fractint IFS format to @fh.
 **/
void
ifs_write_ifs(const IFS *ifs,
              FILE *fh)
{
    size_t i;

    fprintf(fh, "%s {\n", ifs->name);
    for (i = 0; i < ifs->n; i++) {
        IFSTrans *t = ifs->t + i;

        fprintf(fh, "  %f %f %f %f %f %f %f\n",
                t->xx, -t->xy, -t->yx, t->yy,
                t->x, -t->y,
                t->p);
    }
    fprintf(fh, "}\n\n");
}

/**
 * ifs_export_ifscompose:
 * @ifs: An iterated function system.
 * @fh: A file handle open for writing to write the converted IFS to.
 *
 * Write @ifs in Gimp IFS Compose format to @fh.
 **/
void
ifs_export_ifscompose(const IFS *ifs,
                      FILE *fh)
{
    size_t i;
    PointXY center;
    IFSCTrans it;

    center = ifs_find_center(ifs);

    /* header */
    fprintf(fh,
            "iterations %lu\n"
            "max_memory 4096\n"
            "subdivide 3\n"
            "radius 0.618034\n"
            "aspect_ratio 1.0\n"
            "center_x %f\n"
            "center_y %f\n",
            20000LU*ifs->n, center.x, center.y);

    /* transformations */
    for (i = 0; i < ifs->n; i++) {
        ifs_trans_to_ifscompose(ifs->t + i, center, &it);
        fprintf(fh,
                "element {\n"
                "  x %f\n"
                "  y %f\n"
                "  theta %f\n"
                "  scale %.16f\n"
                "  asym %.16f\n"
                "  shear %f\n"
                "  flip %d\n"
                "  red_color { 0.5, 0.0, 0.0 }\n"
                "  green_color { 0.0, 0.5, 0.0 }\n"
                "  blue_color { 0.0, 0.0, 0.5 }\n"
                "  black_color { 0.0, 0.0, 0.0 }\n"
                "  target_color { 0.0, 0.0, 0.0 }\n"
                "  hue_scale 0.5\n"
                "  value_scale 0.5\n"
                "  simple_color 1\n"
                "  prob %f\n"
                "}\n",
                it.x, it.y,
                it.theta, it.scale, it.asym, it.shear,
                it.flip, it.prob*ifs->n);
    }
}

/**
 * ifs_read_ifscompose:
 * @fh: A stdio filehandle open for read.
 * @center: Where the IFS `center' shoud be stored.
 * @ntrans: Where the number of trasformations in the IFS should be stored.
 *
 * Read a GIMP IFS Compose style IFS from file.
 *
 * Returns: A newly allocated GIMP IFS Compose style iterated function system
 *          description.  If the read fails, %NULL is returned.
 **/
IFSCTrans*
ifs_read_ifscompose(FILE *fh,
                    PointXY *center,
                    size_t *ntrans)
{
    IFSCTrans *it, *ifsc = NULL;
    IFSFilePos *pos = NULL;
    const char *p;
    double x;
    int keyword;

    pos = (IFSFilePos*)calloc(1, sizeof(IFSFilePos));
    /* global */
    center->x = center->y = 0.5;
    while (true) {
        keyword = get_ifsc_keyword(fh, pos);
        if (keyword == IFSC_KEY_ELEMENT)
            break;
        switch (keyword) {
            case IFSC_KEY_CENTER_X:
            if (!get_double(fh, pos, &center->x))
                goto fail;
            break;

            case IFSC_KEY_CENTER_Y:
            if (!get_double(fh, pos, &center->y))
                goto fail;
            break;

            case IFSC_KEY_ITERATIONS:
            case IFSC_KEY_MAX_MEMORY:
            case IFSC_KEY_SUBDIVIDE:
            case IFSC_KEY_RADIUS:
            case IFSC_KEY_ASPECT_RATIO:
            if (!get_double(fh, pos, &x))
                goto fail;
            continue;
            break;

            default:
            goto fail;
            break;
        }
    }

    *ntrans = 0;
    while (keyword == IFSC_KEY_ELEMENT
           || ((keyword = get_ifsc_keyword(fh, pos)) == IFSC_KEY_ELEMENT)) {
        (*ntrans)++;
        ifsc = (IFSCTrans*)realloc(ifsc, (*ntrans)*sizeof(IFSCTrans));
        it = ifsc + (*ntrans-1);
        it->flip = 0;
        it->shear = it->theta = it->x = it->y = 0.0;
        it->scale = it->asym = it->prob = 1.0;
        p = get_token(fh, pos);
        if (!p || *p != '{' || *(p+1) != '\0')
            goto fail;
        while ((keyword = get_ifsc_keyword(fh, pos)) != IFSC_LAST_KEYWORD) {
            switch (keyword) {
                case IFSC_KEY_X:
                if (!get_double(fh, pos, &it->x))
                    goto fail;
                break;

                case IFSC_KEY_Y:
                if (!get_double(fh, pos, &it->y))
                    goto fail;
                break;

                case IFSC_KEY_THETA:
                if (!get_double(fh, pos, &it->theta))
                    goto fail;
                break;

                case IFSC_KEY_SCALE:
                if (!get_double(fh, pos, &it->scale))
                    goto fail;
                break;

                case IFSC_KEY_ASYM:
                if (!get_double(fh, pos, &it->asym))
                    goto fail;
                break;

                case IFSC_KEY_SHEAR:
                if (!get_double(fh, pos, &it->shear))
                    goto fail;
                break;

                case IFSC_KEY_PROB:
                if (!get_double(fh, pos, &it->prob))
                    goto fail;
                break;

                case IFSC_KEY_FLIP:
                if (!get_double(fh, pos, &x))
                    goto fail;
                it->flip = x > 0.0;
                break;

                /* ignore these */
                case IFSC_KEY_HUE_SCALE:
                case IFSC_KEY_VALUE_SCALE:
                case IFSC_KEY_SIMPLE_COLOR:
                if (!get_double(fh, pos, &x))
                    goto fail;
                break;

                /* ignore these */
                case IFSC_KEY_RED_COLOR:
                case IFSC_KEY_GREEN_COLOR:
                case IFSC_KEY_BLUE_COLOR:
                case IFSC_KEY_BLACK_COLOR:
                case IFSC_KEY_TARGET_COLOR:
                p = get_token(fh, pos);
                if (!p || *p != '{' || *(p+1) != '\0')
                    goto fail;
                if (!get_and_ditch_tuple(3, fh, pos)
                    || !(p = get_token(fh, pos)))
                    goto fail;
                if (*p != '}' || *(p+1) != '\0')
                    goto fail;
                break;

                default:
                goto fail;
                break;
            }
        }
    }
    ifs_file_pos_free(pos);

    return ifsc;

fail:
    ifs_file_pos_free(pos);
    free(ifsc);
    *ntrans = 0;
    return NULL;
}

static int
get_ifsc_keyword(FILE *fh,
                 IFSFilePos *pos)
{
    const char *p;
    int i;

    if (!(p = get_token(fh, pos)))
        return IFSC_LAST_KEYWORD;
    /* FIXME: inefficient */
    for (i = 0; i < IFSC_LAST_KEYWORD; i++) {
        if (!strcmp(p, IFSC_KEYWORDS[i].name))
            return i;
    }

    return IFSC_LAST_KEYWORD;
}

static bool
get_double(FILE *fh,
           IFSFilePos *pos,
           double *dbl)
{
    const char *p;

    if (!(p = get_token(fh, pos)))
        return false;
    *dbl = atof(p);

    return true;
}

static bool
get_and_ditch_tuple(size_t n,
                    FILE *fh,
                    IFSFilePos *pos)
{
    const char *p = NULL, *end;
    size_t i = 0;

    end = (char*)&i;
    while (i < n) {
        if (!*end) {
            if (!(p = end = get_token(fh, pos)))
                return false;
        }

        strtod(p, (char**)&end);
        if (end == p)
            return false;

        i++;
        if (i == n)
            break;

        if (!*end) {
            if (!(p = end = get_token(fh, pos)))
                return false;
        }
        if (*end != ',')
            return false;
        p = ++end;
    }

    return true;
}

/* get next space-delimited token from file fh */
static const char*
get_token(FILE *fh,
          IFSFilePos *pos)
{
    ssize_t n;

    do {
        if (!pos->buffer || !pos->tok
            || *pos->tok == ifs_comment1
            || *pos->tok == ifs_comment2) {
            n = getline(&pos->buffer, &pos->len, fh);
            if (n < 0)
                return NULL;
            pos->tok = strtok_r(pos->buffer, ifs_delim, &pos->p);
        }
        else
            pos->tok = strtok_r(NULL, ifs_delim, &pos->p);
    } while (!pos->tok
             || *pos->tok == ifs_comment1
             || *pos->tok == ifs_comment2);

    return pos->tok;
}

static const char*
ifs_file_pos_rewind(IFSFilePos *pos)
{
    pos->tok = strtok_r(pos->buffer, ifs_delim, &pos->p);

    return pos->tok;
}

static void
ifs_file_pos_free(IFSFilePos *pos)
{
    free(pos->buffer);
    free(pos);
}

/**
 * ifs_read_fdesign:
 * @fh: A stdio filehandle open for read (better in binary mode).
 *
 * Read a IFS from FDESIGN TRN file.
 *
 * Returns: A newly allocated iterated function system structure.  If the read
 *          fails, %NULL is returned.
 **/
IFS*
ifs_read_fdesign(FILE *fh)
{
    IFS *ifs;
    uint16_t file_revision, i16;
    size_t i;
    double trn0[6], trn[6];

    /* XXX: this works only on intel */
    fread(&file_revision, sizeof(file_revision), 1, fh);
    fread(&i16, sizeof(i16), 1, fh);

    if (file_revision < 9 || file_revision > 500 || i16 > 33)
        return NULL;

    if (!read_triangle(fh, trn0, file_revision <= 300))
        return NULL;

    ifs = (IFS*)malloc(sizeof(IFS));
    ifs->n = i16;
    ifs->t = (IFSTrans*)malloc(ifs->n*sizeof(IFSTrans));
    ifs->name = NULL;

    for (i = 0; i < ifs->n; i++) {
        if (!read_triangle(fh, trn, file_revision <= 300)) {
            ifs_free(ifs);
            return NULL;
        }
        ifs_trans_from_triangle(trn0, trn, ifs->t + i);
    }
    ifs_normalize_probabilities(ifs);
    ifs_flip_vertically(ifs);
    ifs_reflect(ifs, -M_PI/4);

    return ifs;
}

static bool
read_triangle(FILE *fh,
              double *triangle,
              bool old)
{
    uint16_t i16;
    float flt;
    size_t i;

    /* XXX: this works only on intel */
    if (old) {
        for (i = 0; i < 6; i++) {
            if (!fread(&i16, sizeof(i16), 1, fh))
                return false;
            triangle[i] = i16;
        }
    }
    else {
        for (i = 0; i < 6; i++) {
            if (!fread(&flt, sizeof(float), 1, fh)) {
                return false;
            }
            triangle[i] = flt;
        }
    }

    return true;
}

static void
ifs_trans_from_triangle(const double *trn0,
                        const double *trn,
                        IFSTrans *t)
{
    double det, a, b, c, smin, smax;

    triangle_to_coord(trn0, trn, &t->xx, &t->xy, &t->x);
    triangle_to_coord(trn0, trn + 3, &t->yx, &t->yy, &t->y);

    /* fdesign uses probability proportional to triangle size */
    det = trn[3]*(trn[1] - trn[2])
          + trn[4]*(trn[2] - trn[0])
          + trn[5]*(trn[0] - trn[1]);

    /* if the minimum side length is smaller than 1, assume it is a linear
     * element, though we compute the probability differently anyway */
    a = hypot(trn[4] - trn[3], trn[1] - trn[0]);
    b = hypot(trn[5] - trn[4], trn[2] - trn[1]);
    c = hypot(trn[3] - trn[5], trn[0] - trn[2]);
    smin = (a < b) ? ((a < c) ? a : c) : ((b < c) ? b : c);
    smax = (a > b) ? ((a > c) ? a : c) : ((b > c) ? b : c);
    if (smin <= 1.0 + IFS_EPSILON)
        t->p =  5.0*smax;
    else
        t->p = fabs(0.5*det);
}

static void
triangle_to_coord(const double *t,
                  const double *rhs,
                  double *x, double *y, double *c)
{
    double det;

    det = t[0]*(t[4]-t[5])
          - t[3]*(t[1]-t[2])
          + (t[1]*t[5] - t[4]*t[2]);
    *x = rhs[0]*(t[4]-t[5])
         - t[3]*(rhs[1]-rhs[2])
         + (rhs[1]*t[5] - t[4]*rhs[2]);
    *y = t[0]*(rhs[1]-rhs[2])
         - rhs[0]*(t[1]-t[2])
         + (t[1]*rhs[2] - rhs[1]*t[2]);
    *c = t[0]*(t[4]*rhs[2] - rhs[1]*t[5])
         - t[3]*(t[1]*rhs[2] - rhs[1]*t[2])
         + rhs[0]*(t[1]*t[5] - t[4]*t[2]);

    *x /= det;
    *y /= det;
    *c /= det;
}

/* vim: set cin et ts=4 sw=4 cino=>1s,e0,n0,f0,{0,}0,^0,\:1s,=0,g1s,h0,t0,+1s,c3,(0,u0 : */
