/*
 *  @(#) $Id: ifs.c,v 1.39 2004/05/13 18:11:34 yeti Exp $
 *  Copyright (C) 2004 David Necas (Yeti)
 *  E-mail: yeti@physics.muni.cz.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA
 */

#include "ifsgr.h"
#include <stdlib.h>
#include <math.h>
#include <string.h>

/**
 * ifs_free:
 * @ifs: An Iterated Function System.
 *
 * Completely frees an IFS structure.
 **/
void
ifs_free(IFS *ifs)
{
    free(ifs->name);
    free(ifs->t);
    free(ifs);
}

/**
 * ifs_normalize_probabilities:
 * @ifs: An Iterated Function System.
 *
 * Normalize probabilitites to sum 1.0 keeping their ratios.
 **/
void
ifs_normalize_probabilities(IFS *ifs)
{
    double psum, chksum;
    size_t i;

    psum = 0.0;
    for (i = 0; i < ifs->n; i++)
        psum += ifs->t[i].p;

    chksum = 0.0;
    if (psum == 0.0) {
        psum = 1.0/ifs->n;
        for (i = 0; i < ifs->n; i++) {
            ifs->t[i].p = psum;
            chksum += ifs->t[i].p;
        }
    }
    else {
        for (i = 0; i < ifs->n; i++) {
            ifs->t[i].p /= psum;
            chksum += ifs->t[i].p;
        }
    }

    if (ifs->n && chksum < 1.0)
        ifs->t[ifs->n-1].p += 1.0 - chksum;
}

/**
 * ifs_trans_get_fixed_point:
 * @t: An affine transformation.
 *
 * Returns the fixed point of an affine transformation @t.
 *
 * Returns: The fixed point of @t.  When @t is [almost] a pure translation
 * or reflection, %HUGE_VAL is returned.
 **/
PointXY
ifs_trans_get_fixed_point(const IFSTrans *t)
{
    PointXY fp;
    double D, Dx, Dy;

    D = (t->xx-1.0)*(t->yy-1.0) - t->xy*t->yx;
    if (D < IFS_EPSILON) {
        fp.x = fp.y = HUGE_VAL;
        return fp;
    }
    Dx = t->xy*t->y - (t->yy-1.0)*t->x;
    Dy = t->yx*t->x - (t->xx-1.0)*t->y;
    fp.x = Dx/D;
    fp.y = Dy/D;

    return fp;
}

/**
 * ifs_trans_get_scales:
 * @t: An affine transformation.
 *
 * Find the two pure scaling factors of the R2.S.R1 matrix decomposition of
 * @t.
 *
 * R2, R1 are pure rotations, S is a diagonal matrix.  In other words, it
 * finds the unique (up to pi/2 in the rotations) shear-free decomposition
 * and returns the diagonal of S.
 *
 * The pure scaling factors allow to determine easy
 * whether a transformation is a contraction: the necessary and sufficient
 * condition is that both are smaller than one in absolute value.  However,
 * if you only need to know whether a transformation is a contraction,
 * use ifs_trans_is_contraction().
 *
 * Returns: The pure scaling factors, the x factor is always greater or equal
 *          than the y factor (and particularly the x factor is always
 *          nonnegative).
 **/
PointXY
ifs_trans_get_scales(const IFSTrans *t)
{
    double D2, K2, s1, s2;
    PointXY pq;

    K2 = t->xx*t->xx + t->xy*t->xy + t->yx*t->yx + t->yy*t->yy;
    D2 = 2*(t->xx*t->yy - t->xy*t->yx);
    if (fabs(D2) > K2)
        D2 = copysign(K2, D2);
    s1 = sqrt(K2 + D2);
    s2 = sqrt(K2 - D2);
    pq.x = 0.5*(s1 + s2);
    pq.y = 0.5*(s1 - s2);

    return pq;
}

/**
 * ifs_trans_get_rotations:
 * @t: An affine transformation.
 *
 * Find the two rotations of the R2.S.R1 matrix decomposition of @t.
 *
 * Please see ifs_trans_get_scales() for details.
 *
 * Returns: The angles of the two rotations.  The angle of R1 is in the x
 *          component, the angle of R2 is in the y component.
 **/
PointXY
ifs_trans_get_rotations(const IFSTrans *t)
{
    double Trace, Delta, X, Y, sum, diff;
    PointXY phi;

    Trace = t->xx + t->yy;
    Delta = t->yx - t->xy;
    X = t->xy + t->yx;
    Y = t->xx - t->yy;
    sum = atan2(Delta, Trace);
    if (fabs(X) < IFS_EPSILON && fabs(Y) < IFS_EPSILON) {
        phi.x = sum;
        phi.y = 0.0;
    }
    else {
        diff = atan2(X, Y);
        phi.x = 0.5*(sum - diff);
        phi.y = 0.5*(sum + diff);
    }

    return phi;
}

/**
 * ifs_trans_is_contraction:
 * @t: An affine transformation.
 *
 * Test whether affine transformation @t is a contraction.
 *
 * Returns: True if @t is contraction, false otherwise.
 **/
bool
ifs_trans_is_contraction(const IFSTrans *t)
{
    double D, K2;

    K2 = t->xx*t->xx + t->xy*t->xy + t->yx*t->yx + t->yy*t->yy;
    if (K2 >= 2.0)
        return false;

    D = t->xx*t->yy - t->xy*t->yx;
    return K2 - D*D < 1.0;
}

/**
 * ifs_trans_rsr_compose:
 * @pq: Pure scaling factors from R2.S.R1 decomposition.
 * @phi: Rotation angles from R2.S.R1 decomposition.
 * @t: A affine transformation to modify, or %NULL (a new transformation
 *     is allocated then).
 *
 * Composes transformation matrix back from R2.S.R1 decomposition @pq, @phi.
 *
 * The translation and probability components are not defined.
 *
 * Returns: Either @t itself (when non-%NULL), or a newly allocated
 *          transformation that should be freed when no longer needed.
 **/
IFSTrans*
ifs_trans_rsr_compose(PointXY pq,
                      PointXY phi,
                      IFSTrans *t)
{
    double c1, s1, c2, s2;

    if (!t) {
        t = (IFSTrans*)malloc(sizeof(IFSTrans));
        t->x = t->y = t->p = 0.0;
    }

    c1 = cos(phi.x);
    s1 = sin(phi.x);
    c2 = cos(phi.y);
    s2 = sin(phi.y);
    t->xx =  pq.x*c1*c2 - pq.y*s1*s2;
    t->xy = -pq.x*s1*c2 - pq.y*c1*s2;
    t->yx =  pq.x*c1*s2 + pq.y*s1*c2;
    t->yy = -pq.x*s1*s2 + pq.y*c1*c2;

    return t;
}

/**
 * ifs_find_center:
 * @ifs: An Iterated Function System.
 *
 * Find fractal `center'.
 *
 * This is just some point in the convex hull, nothing special.
 * Namely it is the center of fixed points of all transformations, because
 * these are points belonging to the fractal by definition (unless the
 * transformations are dilations).
 *
 * This is the same definition of center as IFS Compose uses, however this
 * function should return something not-so-eccentric even for IFSes containing
 * transformations without fixed points (they are excluded from the
 * computation).
 *
 * Returns: The fractal center.
 **/
PointXY
ifs_find_center(const IFS *ifs)
{
    PointXY center;
    size_t i, n;

    n = 0;
    center.x = center.y = 0.0;
    for (i = 0; i < ifs->n; i++) {
        PointXY fp = ifs_trans_get_fixed_point(ifs->t + i);
        if (isinf(fp.x) || isinf(fp.y))
            continue;
        center.x += fp.x;
        center.y += fp.y;
        n++;
    }

    center.x /= n;
    center.y /= n;
    return center;
}

/**
 * ifs_trans_scale:
 * @t: An affine transformation.
 * @r: Scale factor.
 *
 * Scale the transformation @t by a factor @r.
 **/
void
ifs_trans_scale(IFSTrans *t,
                double r)
{
    t->x *= r;
    t->y *= r;
}

/**
 * ifs_scale:
 * @ifs: An Iterated Function System.
 * @r: Scale factor.
 *
 * Scale the IFS @ifs by a factor @r.
 **/
void
ifs_scale(IFS *ifs,
          double r)
{
    size_t i;

    for (i = 0; i < ifs->n; i++) {
        IFSTrans *t = ifs->t + i;

        t->x *= r;
        t->y *= r;
    }
}

/**
 * ifs_trans_move:
 * @t: An affine transformation.
 * @shift: A vector.
 *
 * Move the transformation @t by the vector @shift.
 **/
void
ifs_trans_move(IFSTrans *t,
               PointXY shift)
{
    t->x += (1.0 - t->xx)*shift.x - t->xy*shift.y;
    t->y += -t->yx*shift.x + (1.0 - t->yy)*shift.y;
}

/**
 * ifs_move:
 * @ifs: An Iterated Function System.
 * @shift: A vector.
 *
 * Move the IFS @ifs by the vector @shift.
 **/
void
ifs_move(IFS *ifs,
         PointXY shift)
{
    size_t i;

    for (i = 0; i < ifs->n; i++) {
        IFSTrans *t = ifs->t + i;

        t->x += (1.0 - t->xx)*shift.x - t->xy*shift.y;
        t->y += -t->yx*shift.x + (1.0 - t->yy)*shift.y;
    }
}

static inline void
trans_rotate(IFSTrans *t,
             double cp,
             double sp)
{
    IFSTrans tt;
    double x;

    /* vector */
    x    = cp*t->x - sp*t->y;
    t->y = sp*t->x + cp*t->y;
    t->x = x;

    /* matrix */
    tt.xx = cp*t->xx - sp*t->yx;
    tt.xy = cp*t->xy - sp*t->yy;
    tt.yx = sp*t->xx + cp*t->yx;
    tt.yy = sp*t->xy + cp*t->yy;

    t->xx =  tt.xx*cp - tt.xy*sp;
    t->xy =  tt.xx*sp + tt.xy*cp;
    t->yx =  tt.yx*cp - tt.yy*sp;
    t->yy =  tt.yx*sp + tt.yy*cp;
}

/**
 * ifs_trans_rotate:
 * @t: An affine transformation.
 * @phi: An angle (in radians).
 *
 * Rotate the transformation @t by the angle @phi around origin.
 **/
void
ifs_trans_rotate(IFSTrans *t,
                 double phi)
{
    trans_rotate(t, cos(phi), sin(phi));
}

/**
 * ifs_rotate:
 * @ifs: An Iterated Function System.
 * @phi: An angle (in radians).
 *
 * Rotate the IFS @ifs by the angle @phi around origin.
 **/
void
ifs_rotate(IFS *ifs,
           double phi)
{
    size_t i;
    double cp, sp;

    cp = cos(phi);
    sp = sin(phi);
    for (i = 0; i < ifs->n; i++)
        trans_rotate(ifs->t + i, cp, sp);
}

/**
 * ifs_trans_flip_vertically:
 * @t: An affine transformation.
 *
 * Flip the transformation @t about the x axis (i.e., vertically).
 **/
void
ifs_trans_flip_vertically(IFSTrans *t)
{
    t->xy = -t->xy;
    t->yx = -t->yx;
    t->y = -t->y;
}

/**
 * ifs_flip_vertically:
 * @ifs: An Iterated Function System.
 *
 * Flip the IFS @ifs about the x axis (i.e., vertically).
 **/
void
ifs_flip_vertically(IFS *ifs)
{
    size_t i;

    for (i = 0; i < ifs->n; i++) {
        IFSTrans *t = ifs->t + i;

        t->xy = -t->xy;
        t->yx = -t->yx;
        t->y = -t->y;
    }
}

/**
 * ifs_trans_flip_horizontally:
 * @t: An affine transformation.
 *
 * Flip the transformation @t about the y axis (i.e., horizontally).
 **/
void
ifs_trans_flip_horizontally(IFSTrans *t)
{
    t->xy = -t->xy;
    t->yx = -t->yx;
    t->x = -t->x;
}

/**
 * ifs_flip_horizontally:
 * @ifs: An Iterated Function System.
 *
 * Flip the IFS @ifs about the y axis (i.e., horizontally).
 **/
void
ifs_flip_horizontally(IFS *ifs)
{
    size_t i;

    for (i = 0; i < ifs->n; i++) {
        IFSTrans *t = ifs->t + i;

        t->xy = -t->xy;
        t->yx = -t->yx;
        t->x = -t->x;
    }
}

/**
 * ifs_trans_asymmetrize:
 * @t: An affine transformation.
 * @alpha: Asymmetrization amount.
 *
 * Asymmetrize x/y ratio of transformation @t by @alpha squared.
 **/
void
ifs_trans_asymmetrize(IFSTrans *t,
                      double alpha)
{
    double alpha2;

    alpha2 = alpha*alpha;
    t->xy *= alpha2;
    t->yx /= alpha2;
    t->x *= alpha;
    t->y /= alpha;
}

/**
 * ifs_asymmetrize:
 * @ifs: An Iterated Function System.
 * @alpha: Asymmetrization amount.
 *
 * Asymmetrize x/y ratio of IFS @ifs by @alpha squared.
 **/
void
ifs_asymmetrize(IFS *ifs,
                double alpha)
{
    double alpha2;
    size_t i;

    alpha2 = alpha*alpha;
    for (i = 0; i < ifs->n; i++) {
        IFSTrans *t = ifs->t + i;

        t->xy *= alpha2;
        t->yx /= alpha2;
        t->x *= alpha;
        t->y /= alpha;
    }
}

/**
 * ifs_trans_skew:
 * @t: An affine transformation.
 * @k: Skew.
 *
 * Skew x by y*@k in the transformation @t.
 **/
void
ifs_trans_skew(IFSTrans *t,
               double k)
{
    double kyx;

    kyx = k*t->yx;
    t->xy += k*(t->yy - t->xx - kyx);
    t->xx += kyx;
    t->yy -= kyx;

    t->x += k*t->y;
}

/**
 * ifs_skew:
 * @ifs: An Iterated Function System.
 * @k: Skew.
 *
 * Skew x by y*@k in the IFS @ifs.
 **/
void
ifs_skew(IFS *ifs,
         double k)
{
    size_t i;

    for (i = 0; i < ifs->n; i++)
        ifs_trans_skew(ifs->t + i, k);
}

static inline void
trans_reflect(IFSTrans *t,
              double cp,
              double sp)
{
    IFSTrans tt;
    double x;

    /* vector */
    x    = cp*t->x + sp*t->y;
    t->y = sp*t->x - cp*t->y;
    t->x = x;

    /* matrix */
    tt.xx = cp*t->xx + sp*t->yx;
    tt.xy = cp*t->xy + sp*t->yy;
    tt.yx = sp*t->xx - cp*t->yx;
    tt.yy = sp*t->xy - cp*t->yy;

    t->xx =  tt.xx*cp + tt.xy*sp;
    t->xy =  tt.xx*sp - tt.xy*cp;
    t->yx =  tt.yx*cp + tt.yy*sp;
    t->yy =  tt.yx*sp - tt.yy*cp;
}

/**
 * ifs_trans_reflect:
 * @t: An affine transformation.
 * @omega: 
 *
 * Reflect the transformation @t about a general axis passing through origin.
 **/
void
ifs_trans_reflect(IFSTrans *t,
                  double omega)
{
    trans_reflect(t, cos(2*omega), sin(2*omega));
}

/**
 * ifs_reflect:
 * @ifs: An Iterated Function System.
 * @omega: 
 *
 * Reflect the IFS @ifs about a general axis passing through origin.
 **/
void
ifs_reflect(IFS *ifs,
            double omega)
{
    size_t i;
    double cp, sp;

    cp = cos(2*omega);
    sp = sin(2*omega);
    for (i = 0; i < ifs->n; i++)
        trans_reflect(ifs->t + i, cp, sp);
}

/**
 * ifs_trans_transform:
 * @t: An affine transformation to be transfromed.
 * @tt: An affine transformation.
 *
 * Transform @t by @tt.
 *
 * Does not make sense when @tt is singular.
 **/
void
ifs_trans_transform(IFSTrans *t,
                    const IFSTrans *tt)
{
    IFSTrans tmp;
    double det;

    det = tt->xx*tt->yy - tt->xy*tt->yx;

    tmp.xx =  (t->xx*tt->yy - t->xy*tt->yx)/det;
    tmp.xy = -(t->xx*tt->xy - t->xy*tt->xx)/det;
    tmp.yx =  (t->yx*tt->yy - t->yy*tt->yx)/det;
    tmp.yy = -(t->yx*tt->xy - t->yy*tt->xx)/det;

    t->xx = tt->xx*tmp.xx + tt->xy*tmp.yx;
    t->xy = tt->xx*tmp.xy + tt->xy*tmp.yy;
    t->yx = tt->yx*tmp.xx + tt->yy*tmp.yx;
    t->yy = tt->yx*tmp.xy + tt->yy*tmp.yy;
    /* now t is tt.t.tt^-1 */

    tmp.x = (1 - t->xx)*tt->x - t->xy*tt->y + tt->xx*t->x + tt->xy*t->y;
    tmp.y = -t->yx*tt->x + (1 - t->yy)*tt->y + tt->yx*t->x + tt->yy*t->y;

    t->x = tmp.x;
    t->y = tmp.y;
}

/**
 * ifs_transform:
 * @ifs: An Iterated Function System to be transformed.
 * @tt: An affine transformation.
 *
 * Transform @ifs by @tt.
 *
 * Does not make sense when @tt is singular.
 **/
void
ifs_transform(IFS *ifs,
              const IFSTrans *tt)
{
    size_t i;

    for (i = 0; i < ifs->n; i++)
        ifs_trans_transform(ifs->t + i, tt);
}

/**
 * ifs_trans_compose:
 * @t1: Left (second) transformation.
 * @t2: Right (first) transformation.
 * @t: Where the composition should be stored, or %NULL to allocate a new
 *     transformation.
 *
 * Compose two affine transformations @t1.@t2 = @t.
 *
 * Returns: Either @t itself (when non-%NULL), or a newly allocated
 *          transformation that should be freed when no longer needed.
 **/
IFSTrans*
ifs_trans_compose(const IFSTrans *t1,
                  const IFSTrans *t2,
                  IFSTrans *t)
{
    if (!t)
        t = (IFSTrans*)malloc(sizeof(IFSTrans));

    t->xx = t1->xx*t2->xx + t1->xy*t2->yx;
    t->xy = t1->xx*t2->xy + t1->xy*t2->yy;
    t->yx = t1->yx*t2->xx + t1->yy*t2->yx;
    t->yy = t1->yx*t2->xy + t1->yy*t2->yy;

    t->x = t1->xx*t2->x + t1->xy*t2->y + t1->x;
    t->y = t1->yx*t2->x + t1->yy*t2->y + t1->y;

    return t;
}

/**
 * ifs_fix_transforms:
 * @ifs: An Iterated Function System.
 * @min: The minimum bounding coordinates.
 * @max: The maximum bounding coordinates.
 * @ratio: The requested size, may be different from 1, but the fractal is
 *         always centered on [0.5,0.5].
 *
 * Transforms the transformations to fit the fractal to [0,1]x[0,1] square.
 **/
void
ifs_fix_transforms(IFS *ifs,
                   PointXY min,
                   PointXY max,
                   double ratio)
{
    double r;
    PointXY shift;

    /* scale */
    r = max.x - min.x;
    if (max.y - min.y > r)
        r = max.y - min.y;
    r = ratio/r;
    ifs_scale(ifs, r);

    min.x *= r;
    min.y *= r;
    max.x *= r;
    max.y *= r;

    /* move */
    shift.x = -0.5*(min.x + max.x - 1.0);
    shift.y = -0.5*(min.y + max.y - 1.0);
    ifs_move(ifs, shift);
}

/**
 * ifs_trans_init_identity:
 * @t: A transformation to be initialized to identity, or %NULL to allocate
 *     a new one.
 *
 * Initializes a transformation to identity, eventually creating it.
 *
 * Transformation probability is not set.
 *
 * Returns: @t if it isn't %NULL; a newly allocated transformation otherwise.
 **/
IFSTrans*
ifs_trans_init_identity(IFSTrans *t)
{
    if (!t)
        t = (IFSTrans*)malloc(sizeof(IFSTrans));
    t->x = t->y = t->xy = t->yx = 0.0;
    t->xx = t->yy = 1.0;

    return t;
}

/**
 * ifs_trans_init_rotation:
 * @t: A transformation to be initialized to rotation, or %NULL to allocate
 *     a new one.
 * @phi: The rotation angle.
 *
 * Initializes a transformation to rotation by @phi counter-clokwise,
 * eventually creating it.
 *
 * Transformation probability is not set.
 *
 * Returns: @t if it isn't %NULL; a newly allocated transformation otherwise.
 **/
IFSTrans*
ifs_trans_init_rotation(IFSTrans *t,
                        double phi)
{
    if (!t)
        t = (IFSTrans*)malloc(sizeof(IFSTrans));
    t->x = t->y = 0.0;
    t->xx = t->yy = cos(phi);
    t->xy = -sin(phi);
    t->yx = -t->xy;

    return t;
}

/**
 * ifs_trans_init_scaling:
 * @t: A transformation to be initialized to scaling, or %NULL to allocate
 *     a new one.
 * @sx: Scaling in x direction.
 * @sy: Scaling in y direction.
 *
 * Initializes a transformation to scaling by (@sx, @sy) eventually creating
 * it.
 *
 * Transformation probability is not set.
 *
 * Returns: @t if it isn't %NULL; a newly allocated transformation otherwise.
 **/
IFSTrans*
ifs_trans_init_scaling(IFSTrans *t,
                       double sx,
                       double sy)
{
    if (!t)
        t = (IFSTrans*)malloc(sizeof(IFSTrans));
    t->x = t->y = t->xy = t->yx = 0.0;
    t->xx = sx;
    t->yy = sy;

    return t;
}

/**
 * ifs_trans_init_skew:
 * @t: A transformation to be initialized to skew, or %NULL to allocate
 *     a new one.
 * @k: The skew.
 *
 * Initializes a transformation skew @k eventually creating it.
 *
 * Transformation probability is not set.
 *
 * Returns: @t if it isn't %NULL; a newly allocated transformation otherwise.
 **/
IFSTrans*
ifs_trans_init_skew(IFSTrans *t,
                    double k)
{
    if (!t)
        t = (IFSTrans*)malloc(sizeof(IFSTrans));
    t->x = t->y = t->yx = 0.0;
    t->xx = t->yy = 1.0;
    t->xy = k;

    return t;
}

/*********** Documentation ***************************************************/
/**
 * PointXY:
 * @x: X coordinate.
 * @y: Y coordinate.
 *
 * Cartesian coordinates of a point in plane.
 **/

/**
 * IFSTrans:
 * @x: Shift in x coordinate.
 * @y: Shift in y coordinate.
 * @xx: The x-to-x matrix element.
 * @xy: The y-to-x matrix element.
 * @yx: The x-to-y matrix element.
 * @yy: The y-to-y matrix element.
 * @p: The probability associated with this transformation.
 *
 * Structure representing an affine two-dimensional transformation.
 **/

/**
 * IFS:
 * @name: IFS name.
 * @n: The number of transformations (size of @t).
 * @t: Array of @n affine transformations.
 *
 * Structure representing an iterated function system.
 **/

/**
 * IFSCTrans:
 * @x: Shift in x coordinate.
 * @y: Shift in y coordinate.
 * @theta: Rotation (in radians).
 * @scale: Symmetric scaling factor.
 * @asym: Asymmetry (asymmetrization factor).
 * @shear: Shear (skew) coefficient.
 * @flip: Nonzero if the transformation involves a flip.
 * @prob: The probability associated with this transformation.
 *
 * Structure describing an iterated function system in GIMP IFS Compose style.
 *
 * See the TeX files in source distribution for exact meaning and discussion
 * of the parameters.
 **/

/**
 * IFS_EPSILON:
 *
 * A small number (much smaller than 1 and much larger than machine epsilon).
 **/

/* vim: set cin et ts=4 sw=4 cino=>1s,e0,n0,f0,{0,}0,^0,\:1s,=0,g1s,h0,t0,+1s,c3,(0,u0 : */
