/*
 *  @(#) $Id: ifsc.c,v 1.1 2004/03/28 00:07:42 yeti Exp $
 *  Copyright (C) 2004 David Necas (Yeti)
 *  E-mail: yeti@physics.muni.cz.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA
 */

#include "ifsgr.h"
#include <stdlib.h>
#include <math.h>
#include <string.h>

/**
 * ifs_trans_to_ifscompose:
 * @t: An affine transformation.
 * @center: IFS Compose fractal center.
 * @ifsct: IFS Compose affine transformation to modify, or %NULL to allocate
 *         a new one.
 *
 * Convert an affine transformation @t to IFC Compose affine transformation.
 *
 * Returns: @ifsct itself (if it wasn't %NULL), or a newly allocated
 *          transformation that should be freed when no longer needed.
 **/
IFSCTrans*
ifs_trans_to_ifscompose(const IFSTrans *t,
                        PointXY center,
                        IFSCTrans *ifsct)
{
    PointXY pq, phi;
    IFSTrans tt;
    double det, scale, ct, st, sign;

    tt = *t;
    if (!ifsct)
        ifsct = (IFSCTrans*)malloc(sizeof(IFSCTrans));

    ifsct->x = tt.x + tt.xx*center.x + tt.xy*center.y;
    ifsct->y = tt.y + tt.yx*center.x + tt.yy*center.y;
    det = tt.xx*tt.yy - tt.yx*tt.xy;

    /* convert absolute probability to determinant-relative, using the same
     * 0.01 line threshold as IFS Compose does. */
    ifsct->prob = tt.p / (fabs(det) < 0.01 ? 0.01 : fabs(det));

    /* [almost] singular transformation */
    if (fabs(det) < IFS_EPSILON) {
        pq = ifs_trans_get_scales(&tt);
        phi = ifs_trans_get_rotations(&tt);
        /* ->point scaling, change to scaling to an `almost point' */
        if (fabs(pq.x) < sqrt(IFS_EPSILON))
            pq.x = sqrt(IFS_EPSILON);
        /* otherwise ->line scaling, change to scaling to an `almost line' */
        pq.y = IFS_EPSILON/pq.x;
        /* reconstruct */
        ifs_trans_rsr_compose(pq, phi, &tt);
        det = tt.xx*tt.yy - tt.yx*tt.xy;
    }

    /* now |det| >= IFS_EPSILON, albeit artificially */
    if (det >= 0) {
        scale = ifsct->scale = sqrt(det);
        sign = 1;
        ifsct->flip = 0;
    }
    else {
        scale = ifsct->scale = sqrt(-det);
        sign = -1;
        ifsct->flip = 1;
    }

    ifsct->theta = atan2(-tt.yx, tt.xx);

    st = sin(ifsct->theta);
    ct = cos(ifsct->theta);

    ifsct->asym = ifsct->scale/(st*(tt.xx + tt.xy) + ct*(tt.yx + tt.yy));
    ifsct->shear = (ct*(tt.xx + tt.xy) - st*(tt.yx + tt.yy))/ifsct->scale;
    ifsct->shear = ifsct->shear/sign - ifsct->asym;
    if (fabs(ifsct->shear) < IFS_EPSILON)
        ifsct->shear = 0.0;

    /* IFS Compose GUI can't handle scale == 1.0 */
    if (ifsct->scale == 1.0)
        ifsct->scale = 0.9999;

    return ifsct;
}

/**
 * ifs_trans_from_ifscompose:
 * @ifsct: A IFS Compose affine transformation.
 * @center: IFS center, as used in IFS Compose.
 * @t: An affine transformation to recompute from @ifsct (or %NULL to allocate
 *     a new one).
 *
 * Convert an IFS Compose transformation to #IFSTrans.
 *
 * Returns: @t itself, or (if it was %NULL) a newly allocated transformation
 *          that should be freed when no longer needed.
 **/
IFSTrans*
ifs_trans_from_ifscompose(const IFSCTrans *ifsct,
                          PointXY center,
                          IFSTrans *t)
{
    double det, ct, st, sign;

    if (!t)
        t = (IFSTrans*)malloc(sizeof(IFSTrans));

    /* FIXME: maybe we should detect pseudo-singular transforms and fix them
     * back to really singular?  (it's impossible to create something with
     * determinant ~~ 10^-9 in the IFS Compose GUI) */
    det = ifsct->scale*ifsct->scale;
    t->p = ifsct->prob * (det < 0.01 ? 0.01 : det);

    sign = ifsct->flip ? -1.0 : 1.0;
    ct = cos(ifsct->theta);
    st = sin(ifsct->theta);

    t->xx = t->xy = t->yx = t->yy = ifsct->scale;
    t->xx *= sign*ifsct->asym*ct;
    t->xy *= sign*ifsct->shear*ct + st/ifsct->asym;
    t->yx *= -sign*ifsct->asym*st;
    t->yy *= -sign*ifsct->shear*st + ct/ifsct->asym;

    t->x = -t->xx*center.x + -t->xy*center.y + ifsct->x;
    t->y = -t->yx*center.x + -t->yy*center.y + ifsct->y;

    return t;
}

/**
 * ifs_from_ifscompose:
 * @ifsc: An array of IFS Compose style transformations.
 * @ntrans: The number of items in @ifsc.
 * @center: IFS center, as used in IFS Compose.
 *
 * Convert complete fractal from IFS Compose style ot #IFS.
 *
 * Returns: A newly allocated #IFS containing the converted fractal.
 **/
IFS*
ifs_from_ifscompose(const IFSCTrans *ifsc,
                    size_t ntrans,
                    PointXY center)
{
    IFS *ifs;
    size_t i;

    ifs = (IFS*)calloc(1, sizeof(IFS));
    ifs->n = ntrans;
    ifs->t = (IFSTrans*)malloc(ntrans*sizeof(IFSTrans));
    ifs->name = NULL;
    for (i = 0; i < ntrans; i++) {
        ifs_trans_from_ifscompose(ifsc + i, center, ifs->t + i);
        ifs_trans_flip_vertically(ifs->t + i);
    }
    ifs_normalize_probabilities(ifs);

    return ifs;
}


