/*
 *  @(#) $Id: ifsgr.c,v 1.46 2004/04/06 09:59:38 yeti Exp $
 *  Copyright (C) 2004 David Necas (Yeti)
 *  E-mail: yeti@physics.muni.cz.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA
 */

#define _GNU_SOURCE
#include "ifsgr.h"
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <getopt.h>
#include <netinet/in.h>

typedef enum {
    IFS_OUTPUT_PGM,
    IFS_OUTPUT_IFSC,
    IFS_OUTPUT_BOXCOUNT,
    IFS_OUTPUT_HISTOGRAM,
    IFS_OUTPUT_PHCOUNT,
    IFS_OUTPUT_FRACTINT
} IFSOutputType;

typedef enum {
    IFS_TWEAK_NONE = 0,
    IFS_TWEAK_ROTATE,
    IFS_TWEAK_ASYM,
    IFS_TWEAK_SKEW,
    IFS_TWEAK_FLIP_X,
    IFS_TWEAK_FLIP_Y
} IFSTweakType;

typedef struct {
    IFSTweakType tweak;
    double amount;
} IFSTweak;

typedef struct {
    char *filename;
    char *ifsname;
} IFSName;

static void      apply_tweaks        (IFS *ifs,
                                      IFSTweak *tweaks);
static bool      render              (IFS **ifsa,
                                      size_t nifs,
                                      const char *ifsname,
                                      const char *filename);
static bool      render_gray8        (IFS *ifs);
static bool      render_gray16       (IFS *ifs);
static bool      render_phcount      (IFS *ifs);
static bool      render_hist         (IFS *ifs);
static bool      convert_to_ifsc     (IFS *ifs);
static bool      convert_to_ifs      (IFS *ifs);
static bool      render_boxcount     (IFS *ifs);
static FILE*     open_output_file    (const char *ifsname,
                                      const char *ext);
static IFS**     read_ifsc_file      (FILE *fh,
                                      const char *filename,
                                      size_t *nifs);
static IFS**     read_trn_file       (FILE *fh,
                                      const char *filename,
                                      size_t *nifs);
static uint32_t* render_int_common   (IFS *ifs,
                                      size_t *m,
                                      double *k);
static IFS*      find_ifs            (IFS **ifsa,
                                      size_t nifs,
                                      const char *ifsname,
                                      const char *filename);
static char*     format_big_number   (uint64_t n);
static IFSName*  process_options     (int argc,
                                      char *argv[]);
static IFSName*  parse_names         (char **args,
                                      int n);
static bool      get_int             (const char *argument,
                                      const char *name,
                                      size_t *value,
                                      size_t min, size_t max);
static bool      get_dbl             (const char *argument,
                                      const char *name,
                                      double *value,
                                      double min, double max);
static bool      get_depth           (const char *argument,
                                      size_t *depth);
static IFSTweak* get_tweak           (const char *argument,
                                      IFSTweak *tweaks,
                                      int type);
static void      print_help          (int exit_code);
static void      print_version       (void);
static char*     filename_to_ifsname (const char *filename,
                                      const char *ext);
static bool      check_little_endian (void);

static inline int
ifs_strcmp(const char *a,
           const char *b)
{
    if (!a || !b) {
        if (!a && !b)
            return 0;
        if (!a)
            return -1;
        return 1;
    }
    return strcmp(a, b);
}

__attribute__((unused))
static struct timeval tv0, tv1;

#define time_start() gettimeofday(&tv0, NULL)
#define time_stop() gettimeofday(&tv1, NULL)
#define elapsed_time() ((tv1.tv_sec - tv0.tv_sec) \
                        + (1e-6*tv1.tv_usec - 1e-6*tv0.tv_usec))

static const size_t ntab = IFS_PROBSTACK;

static bool little_endian = true;
static IFSOutputType outtype = IFS_OUTPUT_PGM;
static IFSTweak *tweaklist = NULL;
static double gammacor = 1.0;
static double quality = 3.0;
static double border = 0.04;
static size_t size = 1024;
static size_t bit_depth = 8;
static int verbose = 0;
static char *outname = NULL;

static bool output_is_dir = false;

int
main(int argc, char *argv[])
{
    IFSName *names, *first;
    IFS **ifsa = NULL;
    size_t i, j, nifs = 0;
    FILE *fh;
    bool ok = true;
    struct stat st;

    little_endian = check_little_endian();
    outname = strdup(".");
    names = process_options(argc, argv);
    stat(outname, &st);
    output_is_dir = S_ISDIR(st.st_mode);
    ifs_taus_seed();

    i = 0;
    first = names;
    while (names[i].filename) {
        IFSName *nm = names + i;

        if (!i || strcmp(nm->filename, first->filename)) {
            for (j = 0; j < nifs; j++)
                ifs_free(ifsa[j]);
            free(ifsa);
            first = nm;
            if (*first->filename) {
                if (!(fh = fopen(first->filename, "r"))) {
                    fprintf(stderr, "%s: Cannot read `%s': %s\n",
                            PACKAGE, first->filename, strerror(errno));
                    while (names[i].filename
                           && !strcmp(names[i].filename, first->filename))
                        i++;
                    ok = false;
                    continue;
                }
            }
            else
                fh = stdin;

            /* check for special file extensions signifying IFSC or TRN,
             * everything else is assumed to be IFS */
            j = strlen(first->filename);
            if (j > 5 && strcmp(first->filename + j-5, ".ifsc") == 0)
                ifsa = read_ifsc_file(fh, first->filename, &nifs);
            else if (j > 4 && strcmp(first->filename + j-4, ".trn") == 0)
                ifsa = read_trn_file(fh, first->filename, &nifs);
            else
                ifsa = ifs_read_file(fh, &nifs);
            if (*first->filename)
                fclose(fh);

            for (j = 0; j < nifs; j++) {
                if (tweaklist)
                    apply_tweaks(ifsa[j], tweaklist);
                ifs_flip_vertically(ifsa[j]);
            }
            if (first->ifsname)
                ok &= render(ifsa, nifs, first->ifsname, nm->filename);
            else {
                for (j = 0; j < nifs; j++)
                    ok &= render(ifsa, nifs, ifsa[j]->name, nm->filename);
            }
        }
        else {
            if (first->ifsname)
                ok &= render(ifsa, nifs, nm->ifsname, nm->filename);
            else
                ok &= (find_ifs(ifsa, nifs, nm->ifsname, nm->filename) != NULL);
        }
        i++;
    }

    return !ok;
}

static void
apply_tweaks(IFS *ifs,
             IFSTweak *tweaks)
{
    size_t i;

    for (i = 0; tweaks[i].tweak; i++) {
        IFSTweak *tw = tweaks + i;

        switch (tw->tweak) {
            case IFS_TWEAK_ROTATE: ifs_rotate(ifs, tw->amount); break;
            case IFS_TWEAK_ASYM: ifs_asymmetrize(ifs, tw->amount); break;
            case IFS_TWEAK_SKEW: ifs_skew(ifs, tw->amount); break;
            case IFS_TWEAK_FLIP_X: ifs_flip_horizontally(ifs); break;
            case IFS_TWEAK_FLIP_Y: ifs_flip_vertically(ifs); break;

            default:
            fprintf(stderr, "%s: Internal error: impossible tweak type %d\n",
                    PACKAGE, tw->tweak);
            exit(EXIT_FAILURE);
            break;
        }
    }
}

static bool
render(IFS **ifsa,
       size_t nifs,
       const char *ifsname,
       const char *filename)
{
    IFS *ifs;

    if (!(ifs = find_ifs(ifsa, nifs, ifsname, filename)))
        return false;

    if (verbose)
        fprintf(stderr, "%s::%s",
                *filename ? filename : "STDIN",
                ifsname ? ifsname : "ALL");

    switch (outtype) {
        case IFS_OUTPUT_PGM:
        if (bit_depth == 8)
            return render_gray8(ifs);
        else if (bit_depth == 16)
            return render_gray16(ifs);
        fprintf(stderr, "%s: Internal error: impossible bit depth %d\n",
                PACKAGE, bit_depth);
        exit(EXIT_FAILURE);
        break;

        case IFS_OUTPUT_PHCOUNT:
        return render_phcount(ifs);
        break;

        case IFS_OUTPUT_BOXCOUNT:
        return render_boxcount(ifs);
        break;

        case IFS_OUTPUT_IFSC:
        return convert_to_ifsc(ifs);
        break;

        case IFS_OUTPUT_FRACTINT:
        return convert_to_ifs(ifs);
        break;

        case IFS_OUTPUT_HISTOGRAM:
        return render_hist(ifs);
        break;

        default:
        fprintf(stderr, "%s: Internal error: impossible output type %d\n",
                PACKAGE, outtype);
        exit(EXIT_FAILURE);
        break;
    }
}

static bool
render_gray8(IFS *ifs)
{
    static unsigned char *pixmap = NULL;
    uint32_t *canvas;
    uint32_t m;
    double k;
    FILE *fh;

    if (!pixmap)
        pixmap = (unsigned char*)malloc(size*size*sizeof(unsigned char));

    canvas = render_int_common(ifs, &m, &k);
    ifs_int2gray8_converg(pixmap, canvas, size, size, m, gammacor*k);

    if (!(fh = open_output_file(ifs->name, "pgm")))
        return false;
    fprintf(fh, "P5 %u %u 255\n", size, size);
    fflush(fh);
    fwrite(pixmap, size*size, sizeof(unsigned char), fh);
    if (fh != stdout)
        fclose(fh);

    return true;
}

static bool
render_gray16(IFS *ifs)
{
    static uint16_t *pixmap = NULL;
    uint32_t *canvas;
    uint32_t m;
    size_t i;
    double k;
    FILE *fh;

    if (!pixmap)
        pixmap = (uint16_t*)malloc(size*size*sizeof(uint16_t));

    canvas = render_int_common(ifs, &m, &k);
    ifs_int2gray16_converg(pixmap, canvas, size, size, m, gammacor*k);
    if (little_endian) {
        for (i = 0; i < size*size; i++)
            pixmap[i] = (pixmap[i] >> 8) | ((pixmap[i] & 0xff) << 8);
    }

    if (!(fh = open_output_file(ifs->name, "pgm")))
        return false;
    fprintf(fh, "P5 %u %u 65535\n", size, size);
    fflush(fh);
    fwrite(pixmap, size*size, sizeof(uint16_t), fh);
    if (fh != stdout)
        fclose(fh);

    return true;
}

static bool
render_phcount(IFS *ifs)
{
    uint32_t *canvas;
    uint32_t m;
    size_t i;
    double k;
    FILE *fh;

    canvas = render_int_common(ifs, &m, &k);
    if (little_endian) {
        for (i = 0; i < size*size; i++)
            canvas[i] = (canvas[i] >> 24)
                        | (((canvas[i] >> 16) & 0xff) << 8)
                        | (((canvas[i] >> 8) & 0xff) << 16)
                        | ((canvas[i] & 0xff) << 24);
    }

    if (!(fh = open_output_file(ifs->name, "pgm32")))
        return false;
    fprintf(fh, "P5 %u %u 4294967295\n", size, size);
    fflush(fh);
    fwrite(canvas, size*size, sizeof(uint32_t), fh);
    if (fh != stdout)
        fclose(fh);

    return true;
}

static bool
render_hist(IFS *ifs)
{
    uint32_t *canvas, *hist;
    uint32_t m;
    size_t i, n;
    double k;
    FILE *fh;

    canvas = render_int_common(ifs, &m, &k);
    hist = ifs_int_get_histogram(size, size, canvas, m);
    n = 0;
    for (i = 1; i < m; i++)
        n += hist[i];

    if (!(fh = open_output_file(ifs->name, "hist")))
        return false;
    for (i = 0; i < m; i++) {
        if (hist[i])
            fprintf(fh, "%u %u %g\n", i, hist[i], (double)hist[i]/n);
    }
    if (fh != stdout)
        fclose(fh);
    free(hist);

    return true;
}

static bool
render_boxcount(IFS *ifs)
{
    uint32_t *canvas, *count;
    uint32_t m;
    size_t n;
    double k, dim;
    FILE *fh;

    canvas = render_int_common(ifs, &m, &k);
    n = (int)sqrt((double)size)+1;
    count = ifs_boxcount(size, size, canvas, n);
    dim = ifs_fit_boxcount_dimension(n, count);
    free(count);

    if (!(fh = open_output_file(ifs->name, "dim")))
        return false;
    fprintf(fh, "%f\n", dim);
    if (fh != stdout)
        fclose(fh);

    return true;
}

static bool
convert_to_ifsc(IFS *ifs)
{
    static size_t *tab = NULL;
    PointXY min, max;

    FILE *fh;

    if (verbose > 1)
        fprintf(stderr, "(%u)", ifs->n);
    tab = ifs_discretize_probability(ifs, ntab, tab);
    ifs_find_range(ifs, tab, 0, &min, &max);
    if (verbose > 1)
        fprintf(stderr, " [%0.2f,%0.2f]x[%0.2f,%0.2f]",
                min.x, max.x, min.y, max.y);
    ifs_fix_transforms(ifs, min, max, 1.0-border);

    if (!(fh = open_output_file(ifs->name, "ifsc")))
        return false;
    ifs_export_ifscompose(ifs, fh);
    if (fh != stdout)
        fclose(fh);

    if (verbose)
        fputc('\n', stderr);

    return true;
}

static bool
convert_to_ifs(IFS *ifs)
{
    static size_t *tab = NULL;
    PointXY min, max;

    FILE *fh;

    if (verbose > 1)
        fprintf(stderr, "(%u)", ifs->n);
    tab = ifs_discretize_probability(ifs, ntab, tab);
    ifs_find_range(ifs, tab, 0, &min, &max);
    if (verbose > 1)
        fprintf(stderr, " [%0.2f,%0.2f]x[%0.2f,%0.2f]",
                min.x, max.x, min.y, max.y);
    ifs_fix_transforms(ifs, min, max, 1.0-border);

    if (!(fh = open_output_file(ifs->name, "ifs")))
        return false;
    ifs_write_ifs(ifs, fh);
    if (fh != stdout)
        fclose(fh);

    if (verbose)
        fputc('\n', stderr);

    return true;
}

static FILE*
open_output_file(const char *ifsname,
                 const char *ext)
{
    static bool do_append = false;
    char *fnm = NULL;
    FILE *fh;

    if (!strcmp(outname, "-"))
        return stdout;

    if (!output_is_dir)
        fnm = strdup(outname);
    else
        asprintf(&fnm, "%s/%s.%s", outname, ifsname, ext);
    if (!(fh = fopen(fnm, do_append ? "a" : "w"))) {
        fprintf(stderr, "%s: cannot open `%s' for writing: %s\n",
                PACKAGE, fnm, strerror(errno));
        return NULL;
    }
    free(fnm);
    if (!output_is_dir)
        do_append = true;

    return fh;
}

static IFS**
read_ifsc_file(FILE *fh,
               const char *filename,
               size_t *nifs)
{
    IFSCTrans *ifsc;
    IFS **ifsa, *ifs;
    PointXY center;
    size_t ntrans;

    ifsc = ifs_read_ifscompose(fh, &center, &ntrans);
    if (!ifsc) {
        fprintf(stderr, "%s: Cannot parse IFSC file `%s'\n",
                PACKAGE, filename);
        *nifs = 0;
        return NULL;
    }
    ifsa = (IFS**)malloc(sizeof(IFS*));
    ifs = ifsa[0] = ifs_from_ifscompose(ifsc, ntrans, center);
    free(ifsc);
    ifs->name = filename_to_ifsname(filename, ".ifsc");
    *nifs = 1;

    return ifsa;
}

static IFS**
read_trn_file(FILE *fh,
              const char *filename,
              size_t *nifs)
{
    IFS **ifsa, *ifs;

    ifs = ifs_read_fdesign(fh);
    if (!ifs) {
        fprintf(stderr, "%s: Cannot parse TRN file `%s'\n",
                PACKAGE, filename);
        *nifs = 0;
        return NULL;
    }
    ifsa = (IFS**)malloc(sizeof(IFS*));
    ifsa[0] = ifs;
    ifs->name = filename_to_ifsname(filename, ".trn");
    *nifs = 1;

    return ifsa;
}

static uint32_t*
render_int_common(IFS *ifs,
                  uint32_t *m,
                  double *k)
{
    static uint32_t *canvas = NULL;
    static size_t *tab = NULL;
    PointXY min, max;
    double dim, measure, avg, ndbl;
    uint64_t niter;

    if (!canvas)
        canvas = (uint32_t*)malloc(size*size*sizeof(uint32_t));

    if (verbose > 1)
        fprintf(stderr, "(%u)", ifs->n);
    tab = ifs_discretize_probability(ifs, ntab, tab);
    ifs_find_range(ifs, tab, 0, &min, &max);
    if (verbose > 1)
        fprintf(stderr, " [%0.2f,%0.2f]x[%0.2f,%0.2f]",
                min.x, max.x, min.y, max.y);
    ifs_fix_transforms(ifs, min, max, 1.0-border);

    ifs_estimate_dimension2(ifs, ntab, tab, &dim, &measure);
    if (verbose > 1)
        fprintf(stderr, " H=%.2f A=%.0f", dim, 4*measure);

    /* the -3.0 is for compatibility with early versions and approximately
     * keeps the interpretation of quality the same after adding measure */
    ndbl = exp((quality-3.0) + dim*log(size))*measure;
    niter = (uint64_t)ndbl;
    if (verbose > 1) {
        char *s = format_big_number(niter);
        fprintf(stderr, " i=%s%s", s, ndbl > 4294967295.0 ? "!" : "");
        free(s);
    }
    if (verbose)
        fputc('\n', stderr);

    ifs_draw_int(ifs, ntab, tab, niter, size, size, canvas);
    *m = ifs_int_get_max(size, size, canvas);
    avg = (double)niter/(size*size)*(1.0-border)*(1.0-border);
    /* XXX: THE empirical madness */
    *k = -0.12/log(1.0 - avg / *m);

    return canvas;
}

static IFS*
find_ifs(IFS **ifsa,
         size_t nifs,
         const char *name,
         const char *filename)
{
    size_t i;

    for (i = 0; i < nifs; i++) {
        if (!strcasecmp(ifsa[i]->name, name))
            return ifsa[i];
    }
    fprintf(stderr, "%s: There is no such IFS like `%s' in `%s'\n",
            PACKAGE, name, *filename ? filename : "STDIN");
    return NULL;
}

static IFSName*
process_options(int argc,
                char *argv[])
{
    static const char *short_options =
        "a:Bd:FGg:HhIk:m:o:Pq:r:s:Vvxy";

    static const struct option long_options[] = {
        { "asymmetrize", required_argument, NULL, 'a' },
        { "box-count", no_argument, NULL, 'B' },
        { "quality", required_argument, NULL, 'q' },
        { "depth", required_argument, NULL, 'd' },
        { "flip-horizontally", no_argument, NULL, 'x' },
        { "flip-vertically", no_argument, NULL, 'y' },
        { "fractint", no_argument, NULL, 'F' },
        { "help", no_argument, NULL, 'h' },
        { "histogram", no_argument, NULL, 'H' },
        { "gamma", required_argument, NULL, 'g' },
        { "ifs-compose", no_argument, NULL, 'I' },
        { "margin", required_argument, NULL, 'm' },
        { "output", required_argument, NULL, 'o' },
        { "pgm", no_argument, NULL, 'G' },
        { "pixel-hit-count", no_argument, NULL, 'P' },
        { "rotate", required_argument, NULL, 'r' },
        { "size", required_argument, NULL, 's' },
        { "skew", required_argument, NULL, 'k' },
        { "verbose", no_argument, NULL, 'V' },
        { "version", no_argument, NULL, 'v' },
        { NULL, 0, NULL, 0 }
    };
    int c;

    opterr = 0;
    while ((c = getopt_long(argc, argv, short_options,
                            long_options, NULL)) != -1) {
        switch (c) {
            case '?': /* Unknown option. */
            fprintf(stderr, "%s: Unknown option -%c%s.\n"
                    "Run `%s --help' to get brief help.\n",
                    PACKAGE, optopt,
                    optopt == '\0' ? " or misspelt/ambiguous long option"
                    : "",
                    PACKAGE);
            exit(EXIT_FAILURE);
            break;

            case ':': /* Missing paramter. */
            fprintf(stderr, "%s: Option -%c requires an argument.\n"
                    "Run `%s --help' to get brief help.\n",
                    PACKAGE, optopt, PACKAGE);
            exit(EXIT_FAILURE);
            break;

            case 'h': print_help(EXIT_SUCCESS); break;
            case 'v': print_version(); break;
            case 'V': verbose++; break;
            case 'B': outtype = IFS_OUTPUT_BOXCOUNT; break;
            case 'F': outtype = IFS_OUTPUT_FRACTINT; break;
            case 'G': outtype = IFS_OUTPUT_PGM; break;
            case 'H': outtype = IFS_OUTPUT_HISTOGRAM; break;
            case 'I': outtype = IFS_OUTPUT_IFSC; break;
            case 'P': outtype = IFS_OUTPUT_PHCOUNT; break;
            case 's': get_int(optarg, "size", &size, 16, 65536); break;
            case 'd': get_depth(optarg, &bit_depth); break;
            case 'g': get_dbl(optarg, "gamma", &gammacor, 0.01, 100); break;
            case 'q': get_dbl(optarg, "quality", &quality, -10.0, 20); break;
            case 'm': get_dbl(optarg, "margin width", &border, 0, 0.999); break;

            case 'r': case 'a': case 'k': case 'x': case 'y':
            tweaklist = get_tweak(optarg, tweaklist, c);
            break;

            case 'o':
            if (outname)
                free(outname);
            outname = strdup(optarg);
            break;

            default:
            fprintf(stderr, "%s: Internal error: unhandled option `-%c'\n",
                    PACKAGE, optopt);
            exit(EXIT_FAILURE);
            break;
        }
    }
    if (optind == argc) {
        fprintf(stderr, "%s: No input files\n"
                "Run `%s --help' to get brief help.\n",
                PACKAGE, PACKAGE);
        exit(EXIT_FAILURE);
    }

    return parse_names(argv + optind, argc - optind);
}

static int
compare_names(const void *a, const void *b)
{
    const IFSName *p = (IFSName*)a;
    const IFSName *q = (IFSName*)b;
    int i;

    i = strcmp(p->filename, q->filename);
    if (i)
        return i;

    return ifs_strcmp(p->ifsname, q->ifsname);
}

static IFSName*
parse_names(char **args,
            int n)
{
    IFSName *names;
    char *p;
    size_t l;
    int i;

    names = (IFSName*)malloc((n+1)*sizeof(IFSName));
    for (i = 0; i < n; i++) {
        p = strstr(args[i], "::");
        if (p) {
            names[i].filename = strndup(args[i], p - args[i]);
            p += 2;
            names[i].ifsname = *p ? strdup(p) : NULL;

            /* check for ifsc */
            l = strlen(names[i].filename);
            if (l > 5 && strcmp(names[i].filename + l-5, ".ifsc") == 0) {
                fprintf(stderr, "%s: Fractal identifiers not supported "
                        "in IFS Compose files (`%s')\n",
                        PACKAGE, args[i]);
                free(names[i].ifsname);
                names[i].ifsname = NULL;
            }
        }
        else {
            names[i].filename = strdup(args[i]);
            names[i].ifsname = NULL;
        }
    }
    qsort(names, n, sizeof(IFSName), compare_names);
    names[n].filename = names[n].ifsname = NULL;

    return names;
}

static bool
get_int(const char *argument,
        const char *name,
        size_t *value,
        size_t min, size_t max)
{
    if (*argument) {
        long int x;
        char *end;

        x = strtol(argument, &end, 0);
        if (!*end) {
            if ((size_t)x >= min && (size_t)x <= max) {
                *value = (size_t)x;
                return true;
            }
            fprintf(stderr, "%s: Ignoring %s value outside of range\n",
                    PACKAGE, name);
            return false;
        }
    }
    fprintf(stderr, "%s: Ignoring invalid %s value\n", PACKAGE, name);
    return false;
}

static bool
get_dbl(const char *argument,
        const char *name,
        double *value,
        double min, double max)
{
    if (*argument) {
        double x;
        char *end;

        x = strtod(argument, &end);
        if (!*end) {
            if (x >= min && x <= max) {
                *value = x;
                return true;
            }
            fprintf(stderr, "%s: Ignoring %s value outside of range\n",
                    PACKAGE, name);
            return false;
        }
    }
    fprintf(stderr, "%s: Ignoring invalid %s value\n", PACKAGE, name);
    return false;
}

static bool
get_depth(const char *argument,
          size_t *depth)
{
    size_t x;

    if (!get_int(argument, "depth", &x, 1, 16))
        return false;
    if (x == 8 || x == 16) {
        *depth = x;
        return true;
    }
    fprintf(stderr, "%s: Only depths 8 and 16 are currently supported\n",
            PACKAGE);
    return false;
}

static IFSTweak*
get_tweak(const char *argument,
          IFSTweak *tweaks,
          int tweak)
{
    static struct {
        int c;
        IFSTweakType tweak;
        const char *name;
        double min;
        double max;
    }
    const tweak_info[] = {
        { 'r', IFS_TWEAK_ROTATE,   "rotation",        -360.0,    360.0    },
        { 'a', IFS_TWEAK_ASYM,     "asymmetrization", 0.001,     1000.0   },
        { 'k', IFS_TWEAK_SKEW,     "skew",            -1000,     1000     },
        { 'x', IFS_TWEAK_FLIP_X,   NULL,              0,         0        },
        { 'y', IFS_TWEAK_FLIP_Y,   NULL,              0,         0        },
    };
    IFSTweak *tw;
    size_t i, j;
    double x = 1.0;

    for (i = 0; i < sizeof(tweak_info)/sizeof(tweak_info[0]); i++) {
        if (tweak_info[i].c == tweak)
            break;
    }
    if (i == sizeof(tweak_info)/sizeof(tweak_info[0])) {
        fprintf(stderr, "%s: Internal error: impossible tweak type `%c'\n",
                PACKAGE, (char)tweak);
        exit(EXIT_FAILURE);
    }

    if (tweak_info[i].name) {
        if (!get_dbl(argument, tweak_info[i].name, &x,
                     tweak_info[i].min, tweak_info[i].max))
            return tweaks;
    }

    if (!tweaks)
        j = 0;
    else {
        for (j = 0; tweaks[j].tweak; j++)
            ;
    }
    tweaks = (IFSTweak*)realloc(tweaks, (j+2)*sizeof(IFSTweak));
    tw = tweaks + j;
    tw->tweak = tweak_info[i].tweak;
    tw->amount = x;
    if (tw->tweak == IFS_TWEAK_ROTATE)
        tw->amount *= M_PI/180;
    if (tw->tweak == IFS_TWEAK_ASYM)
        tw->amount = sqrt(tw->amount);
    tw++;
    tw->tweak = IFS_TWEAK_NONE;

    return tweaks;
}

static void
print_help(int exit_code)
{
    printf("Usage: %s [OPTIONS...] FILE[::IFSNAME] ...\n",
           PACKAGE);
    puts(
"Render Fractint's 2D IFS (Iterated Function System) files to grayscale.\n"
"\n"
"Output:\n"
" -o, --output=PATH         Output file name or directory [from arguments]\n"
" -G, --pgm                 Render to Portable Graymap (.pgm) [default]\n"
" -H, --histogram           Write a pixel hit count historgram (.hist)\n"
" -I, --ifs-compose         Write Gimp IFS Compose spec (.ifsc)\n"
" -F, --fractint            Write Fractint IFS spec (.ifs)\n"
" -B, --box-count           Write box-counting dimension (.dim)\n"
" -P, --pixel-hit-count     Write raw pixel hit counts to pseudo-PGM (.pgm32)\n"
           );
    printf(
"Parameters:\n"
" -s, --size=NUM            Pixmap size [%zu]\n"
" -d, --depth=NUM           Bit depth for PGM (8 or 16) [%zu]\n"
"\n",
           size,
           bit_depth);
    printf(
"Tweaking:\n"
" -q, --quality=NUM         Set quality (the opposite of speed) to NUM [%.1f]\n"
" -g, --gamma=NUM           Set gamma (darkness) correction value to NUM [%.1f]\n"
" -m, --margin=NUM          Set margin width to NUM [%.2f]\n"
"\n",
           quality,
           gammacor,
           border);
    printf(
"Transformations:\n"
" -r, --rotate=NUM          Rotate by NUM degrees CCW\n"
" -a, --asymmetrize=NUM     Asymmetrize x/y aspect ratio by NUM\n"
" -k, --skew=NUM            Skew x by NUM*y\n"
" -x, --flip-horizontally   Flip horizontally\n"
" -y, --flip-vertically     Flip vertically\n"
"\n"
           );
    puts(
"Meta:\n"
" -h, --help                Print this help and terminate\n"
" -v, --version             Print version and terminate\n"
" -V, --verbose             Print progress\n"
"\n"
"Report bugs to David Necas (Yeti) <yeti@physics.muni.cz>\n"
"(please include `ifsgr' in subject)."
        );
    exit(exit_code);
}

static void
print_version(void)
{
    printf("%s %s\n", PACKAGE, VERSION);
    exit(EXIT_SUCCESS);
}

static char*
format_big_number(uint64_t n)
{
    char *s = NULL;

    if (n < 2000ULL)
        asprintf(&s, "%llu", n);
    else if (n < 2000*1000ULL)
        asprintf(&s, "%lluk", n/1000);
    else if (n < 2000*1000000ULL)
        asprintf(&s, "%lluM", n/1000/1000);
    else if (n < 2000*1000000000ULL)
        asprintf(&s, "%lluG", n/1000/1000/1000);
    else
        asprintf(&s, "%lluT", n/1000/1000/1000/1000);

    return s;
}

static char*
filename_to_ifsname(const char *filename,
                    const char *ext)
{
    char *name, *p;
    size_t i, j;

    if (!filename || !strcmp(filename, "-"))
        return strdup("ifs");

    /* XXX: I never know which basename() is used
    name = basename(filename);
    */
    p = strrchr(filename, '/');
    name = strdup(p ? p+1 : filename);
    if (!ext || !*ext)
        return name;

    i = strlen(name);
    j = strlen(ext);
    if (i > j && strcmp(name + i-j, ext) == 0)
        name[i-j] = '\0';

    return name;
}

static bool
check_little_endian(void)
{
    uint16_t a = 0x1234, b;

    b = htons(a);
    return b != a;
}

/* vim: set cin et ts=4 sw=4 cino=>1s,e0,n0,f0,{0,}0,^0,\:1s,=0,g1s,h0,t0,+1s,c3,(0,u0 : */
