/*
 *  @(#) $Id: range.c,v 1.5 2004/05/13 18:11:34 yeti Exp $
 *  Copyright (C) 2004 David Necas (Yeti)
 *  E-mail: yeti@physics.muni.cz.
 *  Copyright (C) 2001-2003 Martin Siler
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA
 */

#include "ifsgr.h"
#include <stdlib.h>
#include <math.h>
#include <string.h>

#if 0
#define IFS_NEW(t, n) ((t*)malloc((n)*sizeof(t)))
#define IFS_NEW0(t, n) ((t*)calloc((n), sizeof(t)))

#define FitSqrtMachEps  1e-4
#define FitEps 1e-16

/* Lower symmetric part indexing */
/* i MUST be greater or equal than j */
#define SLi(a, i, j) a[(i)*((i) + 1)/2 + (j)]

typedef double (*IFSNLFitFunc)(double x,
                               double *param);
typedef void   (*IFSNLFitDFunc)(double x,
                                double *param,
                                double *deriv);

typedef struct _IFSNLFitter IFSNLFitter;

struct _IFSNLFitter {
    IFSNLFitFunc fmarq;  /* fitting function */
    IFSNLFitDFunc dmarq;  /* fitting function derivations */
    size_t maxiter;  /* max number of iterations */
    bool eval;  /* success? */
    double mfi;    /* fitting parameters */
    double mdec;
    double minc;
    double mtol;
};

static double        fit_asymptote          (IFSNLFitter *nlfit,
                                             double *x,
                                             double *z,
                                             size_t midx);
static double        fit_asymptote_ff       (double x,
                                             double *param);
static void          fit_asymptote_df       (double x,
                                             double *param,
                                             double *deriv);
static IFSNLFitter*  ifs_nlfit_new          (IFSNLFitFunc ff,
                                             IFSNLFitDFunc df);
static double        ifs_nlfit_fit          (IFSNLFitter *nlfit,
                                             size_t n_dat,
                                             double *x,
                                             double *y,
                                             size_t n_par,
                                             double *param);
static bool          ifs_choleski_decompose (size_t dim,
                                             double *a);
static void          ifs_choleski_solve     (size_t dim,
                                             double *a,
                                             double *b);
static double        ifs_nlfit_residua      (IFSNLFitter *nlfit,
                                             size_t n_dat,
                                             double *x,
                                             double *y,
                                             double *param,
                                             double *resid);

/* number of segments for intelligent range finder */
enum { RANGE_CSEGMENTS = 48 };

/* number of extremes remembered in range extrapolator */
enum { RANGE_AMSIZE = 160 };


void
ifs_find_range2(const IFS *ifs,
                PointXY *min,
                PointXY *max)
{
    PointXY center = ifs_find_center(ifs);
    size_t i, j, n;
    PointXY *r0, *r1, *r2;
    size_t rsize = RANGE_CSEGMENTS*sizeof(PointXY);
    const double shf = M_PI - M_PI/RANGE_CSEGMENTS;

    r0 = malloc(rsize);
    r1 = malloc(rsize);
    r2 = malloc(rsize);

    for (i = 0; i < RANGE_CSEGMENTS; i++)
        r0[i].x = r0[i].y = r1[i].x = r1[i].y = r2[i].x = r2[i].y = 0.0;

    for (n = 0; n < 100; n++) {
        if (n >= 2) {
            PointXY *rtmp = r2;
            r2 = r1;
            r1 = rtmp;
            memcpy(r1, r0, rsize);
        }
        for (j = 0; j < ifs->n; j++) {
            IFSTrans *t = ifs->t + j;
            for (i = 0; i < RANGE_CSEGMENTS; i++) {
                PointXY *p = r0 + i;
                double x = t->xx*(p->x + center.x)
                    + t->xy*(p->y + center.y)
                    + t->x - center.x;
                double y = t->yx*(p->x + center.x)
                    + t->yy*(p->y + center.y)
                    + t->y - center.y;
                double q = (atan2(y, x) + shf)/(2.0*M_PI);
                size_t idx;
                if (q < 0)
                    q += 1.0;
                idx = floor(RANGE_CSEGMENTS*q);
                if (r0[idx].x*r0[idx].x + r0[idx].y*r0[idx].y < x*x + y*y) {
                    r0[idx].x = x;
                    r0[idx].y = y;
                }
            }
        }
    }

    min->x = min->y = HUGE_VAL;
    max->x = max->y = -HUGE_VAL;
    for (i = 0; i < RANGE_CSEGMENTS; i++) {
        if (r0[i].x < min->x)
            min->x = r0[i].x;
        if (r0[i].y < min->y)
            min->y = r0[i].y;
        if (r0[i].x > max->x)
            max->x = r0[i].x;
        if (r0[i].y > max->y)
            max->y = r0[i].y;
    }
    min->x += center.x;
    max->x += center.x;
    min->y += center.y;
    max->y += center.y;

    free(r0);
    free(r1);
    free(r2);
}
#endif

/**
 * ifs_find_range:
 * @ifs: An Iterated Function System.
 * @tab: Precomputed probability table of size %IFS_PROBSTACK.
 * @niter: The number of iterations to try, when zero some reasonable built-in
 *         default is used.
 * @min: Where the minimum coordinates should be stored.
 * @max: Where the maximum coordinates should be stored.
 *
 * Find the rectangle to which the fractal @ifs fits in.
 *
 * This is done simply by computing a few thousand iterations.
 */
void
ifs_find_range(const IFS *ifs,
               const size_t *tab,
               uint64_t niter,
               PointXY *min,
               PointXY *max)
{
    IFSTrans *t;
    double x, y, xx;
    uint64_t i;

    min->x = min->y = HUGE_VAL;
    max->x = max->y = -HUGE_VAL;
    if (!niter)
        niter = 100000LLU;

    x = y = 0.0;
    /* get to the attractor */
    for (i = 0; i < 12; i++) {
        t = ifs->t + i % ifs->n;
        xx = x;
        x = xx*t->xx + y*t->xy + t->x;
        y = xx*t->yx + y*t->yy + t->y;
    }

    /* find the range */
    for (i = 0; i < niter; i++) {
        t = ifs->t + tab[ifs_taus_get() % IFS_PROBSTACK];
        xx = x;
        x = xx*t->xx + y*t->xy + t->x;
        y = xx*t->yx + y*t->yy + t->y;

        if (x < min->x)
            min->x = x;
        else if (x > max->x)
            max->x = x;
        if (y < min->y)
            min->y = y;
        else if (y > max->y)
            max->y = y;
    }
}

#if 0
/**
 * ifs_find_range_extrapolated:
 * @ifs: An Iterated Function System.
 * @tab: Precomputed probability table of size %IFS_PROBSTACK.
 * @niter: The number of iterations to try, when zero some reasonable built-in
 *         default is used.
 * @min: Where the minimum coordinates should be stored.
 * @max: Where the maximum coordinates should be stored.
 *
 * Find the rectangle to which the fractal @ifs fits in.
 *
 * This is done simply by computing a few thousand iterations.
 */
void
ifs_find_range_extrapolated(const IFS *ifs,
                            const size_t *tab,
                            uint64_t niter,
                            PointXY *min,
                            PointXY *max)
{
    enum { n_tries = 2000 };
    IFSTrans *t;
    double x, y, xx;
    double *xmin, *xmax, *ymin, *ymax, *indep;
    size_t j, i;
    IFSNLFitter *nlfit;
    PointXY tmin, tmax;
    uint64_t k, step;

    if (!niter)
        niter = 10000000LLU;

    x = y = 0.0;
    /* get to the attractor */
    for (i = 0; i < 12; i++) {
        t = ifs->t + i % ifs->n;
        xx = x;
        x = xx*t->xx + y*t->xy + t->x;
        y = xx*t->yx + y*t->yy + t->y;
    }

    /* find the range */
    xmin = IFS_NEW0(double, RANGE_AMSIZE);
    xmax = IFS_NEW0(double, RANGE_AMSIZE);
    ymin = IFS_NEW0(double, RANGE_AMSIZE);
    ymax = IFS_NEW0(double, RANGE_AMSIZE);
    indep = IFS_NEW(double, RANGE_AMSIZE);
    for (i = 0; i < RANGE_AMSIZE; i++)
        indep[i] = i + 1.0;

    step = niter/RANGE_AMSIZE + 1;
    tmin.x = tmin.y = HUGE_VAL;
    tmax.x = tmax.y = -HUGE_VAL;
    for (j = 0; j < n_tries; j++) {
        min->x = min->y = HUGE_VAL;
        max->x = max->y = -HUGE_VAL;
        for (i = 0; i < RANGE_AMSIZE; i++) {
            for (k = 0; k < step; k++) {
                t = ifs->t + tab[ifs_taus_get() % IFS_PROBSTACK];
                xx = x;
                x = xx*t->xx + y*t->xy + t->x;
                y = xx*t->yx + y*t->yy + t->y;

                if (x < min->x)
                    min->x = x;
                if (x > max->x)
                    max->x = x;
                if (y < min->y)
                    min->y = y;
                if (y > max->y)
                    max->y = y;
            }

            xmin[i] += min->x;
            xmax[i] += max->x;
            ymin[i] += min->y;
            ymax[i] += max->y;
        }

        if (min->x < tmin.x)
            tmin.x = min->x;
        if (max->x > tmax.x)
            tmax.x = max->x;
        if (min->y < tmin.y)
            tmin.y = min->y;
        if (max->y > tmax.y)
            tmax.y = max->y;
    }
    for (i = 0; i < RANGE_AMSIZE; i++) {
        xmin[i] /= n_tries;
        xmax[i] /= n_tries;
        ymin[i] /= n_tries;
        ymax[i] /= n_tries;
    }

    fputs("\n", stderr);
    for (i = 0; i < RANGE_AMSIZE; i++)
        fprintf(stderr, "%f %f %f %f %f\n",
                indep[i], xmin[i], xmax[i], ymin[i], ymax[i]);

    fprintf(stderr, "I %f %f %f %f\n", min->x, max->x, min->y, max->y);
    fprintf(stderr, "T %f %f %f %f\n", tmin.x, tmax.x, tmin.y, tmax.y);
    /*fprintf(stderr, "E");*/
    nlfit = ifs_nlfit_new(fit_asymptote_ff, fit_asymptote_df);
    min->x = fit_asymptote(nlfit, indep, xmin, RANGE_AMSIZE);
    max->x = fit_asymptote(nlfit, indep, xmax, RANGE_AMSIZE);
    min->y = fit_asymptote(nlfit, indep, ymin, RANGE_AMSIZE);
    max->y = fit_asymptote(nlfit, indep, ymax, RANGE_AMSIZE);
    free(nlfit);
    free(indep);
    free(ymax);
    free(ymin);
    free(xmax);
    free(xmin);
    fputs("\n", stderr);
    fprintf(stderr, "F %f %f %f %f\n", min->x, max->x, min->y, max->y);
}

static double
fit_asymptote(IFSNLFitter *nlfit,
              double *x,
              double *z,
              size_t midx)
{
    double x0, xn, x2n, z0, zn, z2n, q0, qn, q2n;
    double param[3];  /* 0: scale, 1: exponent, 2: additive constant */

    /* give up with too few data points */
    if (midx < 5)
        return x[midx-1];

    /*
    fprintf(stderr, "== %d\n", midx);
    for (i = 0; i < midx; i++)
        fprintf(stderr, "%f\n", z[i]);
        */

    /* initial parameter guess */
    z0 = z[0];
    z2n = z[midx-1];
    x0 = x[0];
    x2n = x[midx-1];
    if (midx % 2) {
        xn = x[midx/2];
        zn = z[midx/2];
    }
    else {
        xn = (x[midx/2] + x[midx/2 - 1])/2.0;
        zn = (z[midx/2] + z[midx/2 - 1])/2.0;
    }
    /*fprintf(stderr, "z0 = %f; zn = %f; z2n = %f\n", z0, zn, z2n);*/

    q0 = log(x0);
    q0 = 0;
    qn = log(xn);
    q2n = log(x2n);
    param[1] = ((z0 - zn)/(zn - z2n)*(q2n - qn)/qn - 1.0)/q2n;
    param[0] = (z0 - z2n)/(1.0 - 1.0/(1.0 + param[1]*q2n));
    param[2] = (z2n - z0/(1.0 + param[1]*q2n))
               /(1.0 - 1.0/(1.0 + param[1]*q2n));

    fprintf(stderr, "Initial guess: a = %f; b = %f; c = %f\n",
            param[0], param[1], param[2]);
    /*fprintf(stderr, " %f", param[2]);*/

    q0 = ifs_nlfit_fit(nlfit, midx, x, z, 3, param);
    fprintf(stderr, "Final values: a = %f; b = %f; c = %f\n",
            param[0], param[1], param[2]);
    /*fprintf(stderr, "Residual sum of squares: s = %f\n", q);*/

    return param[2];
}

static double
fit_asymptote_ff(double x,
                 double *param)
{
    return param[0]/(1.0 + log(x)*param[1]) + param[2];
}

static void
fit_asymptote_df(double x,
                 double *param,
                 double *deriv)
{
    double q;

    q = 1.0 + param[1]*log(x);
    deriv[0] = 1.0/q;
    deriv[1] = -param[0]*log(x)/(q*q);
    deriv[2] = 1.0;
}

IFSNLFitter*
ifs_nlfit_new(IFSNLFitFunc ff, IFSNLFitDFunc df)
{
    IFSNLFitter *nlfit;

    nlfit = IFS_NEW0(IFSNLFitter, 1);
    nlfit->fmarq = ff;
    nlfit->dmarq = df;
    nlfit->mfi = 1.0;
    nlfit->mdec = 0.4;
    nlfit->minc = 10.0;
    nlfit->mtol = 1e-6;
    nlfit->maxiter = 500;
    nlfit->eval = false;

    return nlfit;
}

double
ifs_nlfit_fit(IFSNLFitter *nlfit,
              size_t n_dat,
              double *x,
              double *y,
              size_t n_par,
              double *param)
{

    double mlambda = 1.0e-4;
    double sumr = HUGE_VAL;
    double *der, *v, *xr, *saveparam, *resid, *a, *save_a;
    bool step1 = true;
    bool end = false;
    size_t i, j, k;
    double sumr1;
    size_t miter = 0;
    size_t covar_size = n_par*(n_par + 1)/2;

    resid = IFS_NEW(double, n_dat);
    sumr1 = ifs_nlfit_residua(nlfit, n_dat, x, y, param, resid);
    sumr = sumr1;

    if (!nlfit->eval) {
        free(resid);
        return -1;
    }

    der = IFS_NEW(double, n_par);
    v = IFS_NEW(double, n_par);
    xr = IFS_NEW(double, n_par);
    saveparam = IFS_NEW(double, n_par);
    a = IFS_NEW(double, covar_size);
    save_a = IFS_NEW(double, covar_size);

    do {
        bool posdef = false;
        bool first_pass = true;
        size_t count = 0;

        if (step1) {
            mlambda *= nlfit->mdec;
            sumr = sumr1;

            for (i = 0; i < covar_size; i++)
                a[i] = 0;
            for (i = 0; i < n_par; i++)
                v[i] = 0;

            for (i = 0; i < n_dat; i++) {
                nlfit->dmarq(x[i], param, der);
                for (j = 0; j < n_par; j++) {
                    size_t q = j*(j + 1)/2;

                    v[j] += der[j] * resid[i];
                    for (k = 0; k <= j; k++)
                        a[q + k] += der[j] * der[k];
                }
            }
            if (nlfit->eval) {
                memcpy(save_a, a, covar_size*sizeof(double));
                memcpy(saveparam, param, n_par*sizeof(double));
            }
            else
                break;
        }
        while (!posdef) {
            if (!first_pass)
                memcpy(a, save_a, covar_size*sizeof(double));
            else
                first_pass = false;

            for (j = 0; j < n_par; j++) {
                size_t q = j*(j + 3)/2;

                a[q] = save_a[q]*(1.0 + mlambda) + nlfit->mfi * mlambda;
                xr[j] = -v[j];
            }

            posdef = ifs_choleski_decompose(n_par, a);
            if (!posdef) {
                mlambda *= nlfit->minc;
                if (mlambda == 0.0)
                    mlambda = nlfit->mtol;
            }
        }
        ifs_choleski_solve(n_par, a, xr);

        for (i = 0; i < n_par; i++) {
            param[i] = saveparam[i] + xr[i];
            if (fabs(param[i] - saveparam[i]) == 0)
                count++;
        }
        if (count == n_par)
            break;

        sumr1 = ifs_nlfit_residua(nlfit, n_dat, x, y, param, resid);
        if ((sumr1 == 0)
            || (miter > 2 && fabs((sumr - sumr1)/sumr1) < FitEps))
            end = true;
        if (!nlfit->eval || sumr1 >= sumr) {
            mlambda *= nlfit->minc;
            if (mlambda == 0.0)
                mlambda = nlfit->mtol;
            step1 = false;
        }
        else
            step1 = true;

        if (++miter >= nlfit->maxiter)
            break;
    } while (!end);

    sumr1 = sumr;

    free(save_a);
    free(a);
    free(resid);
    free(saveparam);
    free(xr);
    free(v);
    free(der);

    return sumr;
}

static double
ifs_nlfit_residua(IFSNLFitter *nlfit,
                  size_t n_dat,
                  double *x,
                  double *y,
                  double *param,
                  double *resid)
{
    double s = 0;
    size_t i;

    nlfit->eval = true;
    for (i = 0; i < n_dat && nlfit->eval; i++) {
        resid[i] = nlfit->fmarq(x[i], param) - y[i];
        s += resid[i]*resid[i];
    }
    return s;
}

static bool
ifs_choleski_decompose(size_t dim, double *a)
{
    size_t i, j, k;
    double s, r;

    for (k = 0; k < dim; k++) {
        /* diagonal element */
        s = SLi(a, k, k);
        for (i = 0; i < k; i++)
            s -= SLi(a, k, i) * SLi(a, k, i);
        if (s <= 0.0)
            return false;
        SLi(a, k, k) = s = sqrt(s);

        /* nondiagonal elements */
        for (j = k+1; j < dim; j++) {
            r = SLi(a, j, k);
            for (i = 0; i < k; i++)
                r -= SLi(a, k, i) * SLi(a, j, i);
            SLi(a, j, k) = r/s;
        }
    }

    return true;
}

static void
ifs_choleski_solve(size_t dim, double *a, double *b)
{
    size_t i, j;

    /* back-substitution with the lower triangular matrix */
    for (j = 0; j < dim; j++) {
        for (i = 0; i < j; i++)
            b[j] -= SLi(a, j, i)*b[i];
        b[j] /= SLi(a, j, j);
    }

    /* back-substitution with the upper triangular matrix */
    for (j = dim-1; j < dim; j--) {
        for (i = j+1; i < dim; i++)
            b[j] -= SLi(a, i, j)*b[i];
        b[j] /= SLi(a, j, j);
    }
}
#endif

/* vim: set cin et ts=4 sw=4 cino=>1s,e0,n0,f0,{0,}0,^0,\:1s,=0,g1s,h0,t0,+1s,c3,(0,u0 : */
