#include "ifsgr.h"
#include <math.h>

#define EPS 1e-13

static IFSTrans identity = { 0.0, 0.0, 1.0, 0.0, 0.0, 1.0, 1.0 };

static int
compare_trans(const IFSTrans *t1, const IFSTrans *t2, const char *test)
{
    bool ok = true;

    if (fabs(t1->x - t2->x) > EPS) {
        ok = false;
        fprintf(stderr, "%s: t1->x = %0.15f <> t2->x = %0.15f\n",
                test, t1->x, t2->x);
    }
    if (fabs(t1->y - t2->y) > EPS) {
        ok = false;
        fprintf(stderr, "%s: t1->y = %0.15f <> t2->y = %0.15f\n",
                test, t1->y, t2->y);
    }
    if (fabs(t1->xx - t2->xx) > EPS) {
        ok = false;
        fprintf(stderr, "%s: t1->xx = %0.15f <> t2->xx = %0.15f\n",
                test, t1->xx, t2->xx);
    }
    if (fabs(t1->xy - t2->xy) > EPS) {
        ok = false;
        fprintf(stderr, "%s: t1->xy = %0.15f <> t2->xy = %0.15f\n",
                test, t1->xy, t2->xy);
    }
    if (fabs(t1->yx - t2->yx) > EPS) {
        ok = false;
        fprintf(stderr, "%s: t1->yx = %0.15f <> t2->yx = %0.15f\n",
                test, t1->yx, t2->yx);
    }
    if (fabs(t1->yy - t2->yy) > EPS) {
        ok = false;
        fprintf(stderr, "%s: t1->yy = %0.15f <> t2->yy = %0.15f\n",
                test, t1->yy, t2->yy);
    }

    return ok ? 1 : 0;
}

static void
print_passed(const char *name,
             int passed,
             int total)
{
    if (total == passed)
        fprintf(stderr, "== %s: all (%d) tests passed\n",
                name, total);
    else
        fprintf(stderr, "== %s: %d of %d tests failed\n",
                name, total-passed, total);

}

/* XXX: this test doesn't test much whether the various functions work
 * correctly, only whether different methods give give consistent results
 * (as the probability of two identical mistakes is of second order). */
static bool
trans_transforms(void)
{
    IFSTrans t = { 0.4, 0.7, 1.0, -0.3, 0.6, -0.25, 1.0 };
    IFSTrans tx = { -10, 10, 0.71, 0.22, 0.45, 1.24, 1.0 };
    IFSTrans tt, t2, test1, test2;
    double scale = 14.0, asym = 4.0, phi = 5.12, skew = 0.443;
    PointXY v = { -1.0, 0.5 };
    PointXY w, pq, rot;
    int total = 0, passed = 0;

    test1 = test2 = t;
    tt = identity;
    tt.xx = scale;
    tt.yy = scale;
    ifs_trans_scale(&test1, scale);
    ifs_trans_transform(&test2, &tt);
    passed += compare_trans(&test1, &test2, "trans_transforms: scale");
    ifs_trans_scale(&test1, 1.0/scale);
    passed += compare_trans(&test1, &t, "trans_transforms: scale (inv)");
    total += 2;

    test1 = test2 = t;
    tt = identity;
    tt.xx = tt.yy = cos(phi);
    tt.yx = sin(phi);
    tt.xy = -tt.yx;
    ifs_trans_rotate(&test1, phi);
    ifs_trans_transform(&test2, &tt);
    passed += compare_trans(&test1, &test2, "trans_transforms: rotate");
    ifs_trans_rotate(&test1, -phi);
    passed += compare_trans(&test1, &t, "trans_transforms: rotate (inv)");
    total += 2;

    test1 = test2 = t;
    tt = identity;
    tt.xy = skew;
    ifs_trans_skew(&test1, skew);
    ifs_trans_transform(&test2, &tt);
    passed += compare_trans(&test1, &test2, "trans_transforms: skew");
    ifs_trans_skew(&test1, -skew);
    passed += compare_trans(&test1, &t, "trans_transforms: skew (inv)");
    total += 2;

    test1 = test2 = t;
    tt = identity;
    tt.xx = asym;
    tt.yy = 1.0/tt.xx;
    ifs_trans_asymmetrize(&test1, asym);
    ifs_trans_transform(&test2, &tt);
    passed += compare_trans(&test1, &test2, "trans_transforms: asymmetrize");
    ifs_trans_asymmetrize(&test1, 1.0/asym);
    passed += compare_trans(&test1, &t, "trans_transforms: asymmetrize (inv)");
    total += 2;

    test1 = test2 = t;
    tt = identity;
    tt.x = v.x;
    tt.y = v.y;
    ifs_trans_move(&test1, v);
    ifs_trans_transform(&test2, &tt);
    passed += compare_trans(&test1, &test2, "trans_transforms: move");
    w.x = -v.x;
    w.y = -v.y;
    ifs_trans_move(&test1, w);
    passed += compare_trans(&test1, &t, "trans_transforms: move (inv)");
    total += 2;

    pq = ifs_trans_get_scales(&tx);
    rot = ifs_trans_get_rotations(&tx);
    test1 = test2 = t;
    ifs_trans_rotate(&test1, rot.x);
    ifs_trans_scale(&test1, sqrt(pq.x*pq.y));
    ifs_trans_asymmetrize(&test1, sqrt(pq.x/pq.y));
    ifs_trans_rotate(&test1, rot.y);
    w.x = tx.x;
    w.y = tx.y;
    ifs_trans_move(&test1, w);
    ifs_trans_transform(&test2, &tx);
    passed += compare_trans(&test1, &test2, "trans_transforms: trans + rsr");
    total += 1;

    t2 = identity;
    t2.x = v.x;
    t2.y = v.y;
    t2.xx = 0.6;
    t2.yx = -1.4;
    ifs_trans_compose(&tx, &t2, &test2);
    ifs_trans_compose(&t, &test2, &test1);
    ifs_trans_compose(&t, &tx, &tt);
    ifs_trans_compose(&tt, &t2, &test2);
    passed += compare_trans(&test1, &test2, "trans_transforms: asociativity");
    total += 1;

    t2 = tt = identity;
    tt.xx = tt.yy = cos(rot.x);
    tt.yx = sin(rot.x);
    tt.xy = -tt.yx;
    ifs_trans_compose(&tt, &t2, &test1);
    tt = identity;
    tt.xx = pq.x;
    tt.yy = pq.y;
    ifs_trans_compose(&tt, &test1, &t2);
    tt = identity;
    tt.xx = tt.yy = cos(rot.y);
    tt.yx = sin(rot.y);
    tt.xy = -tt.yx;
    ifs_trans_compose(&tt, &t2, &test1);
    tt = identity;
    tt.x = tx.x;
    tt.y = tx.y;
    ifs_trans_compose(&tt, &test1, &t2);
    passed += compare_trans(&tx, &t2, "trans_transforms: rsr compose");
    total += 1;

    t2 = tt = identity;
    t2.x = v.x;
    t2.y = v.y;
    tt.xx = tt.yy = cos(rot.x);
    tt.yx = sin(rot.x);
    tt.xy = -tt.yx;
    ifs_trans_compose(&tt, &t2, &test1);
    tt = identity;
    tt.xx = pq.x;
    tt.yy = pq.y;
    ifs_trans_compose(&tt, &test1, &t2);
    tt = identity;
    tt.xx = tt.yy = cos(rot.y);
    tt.yx = sin(rot.y);
    tt.xy = -tt.yx;
    ifs_trans_compose(&tt, &t2, &test1);
    tt = identity;
    tt.x = tx.x;
    tt.y = tx.y;
    ifs_trans_compose(&tt, &test1, &t2);
    test1 = identity;
    test1.x = v.x;
    test1.y = v.y;
    ifs_trans_compose(&tx, &test1, &test2);
    passed += compare_trans(&test2, &t2, "trans_transforms: compose + vec");
    total += 1;

    print_passed("trans_transforms", passed, total);

    return passed == total;
}

int
main(void)
{
    trans_transforms();
}

/* vim: set cin et ts=4 sw=4 cino=>1s,e0,n0,f0,{0,}0,^0,\:1s,=0,g1s,h0,t0,+1s,c3,(0,u0 : */
